;;; Experimental General Dispatch Method

(in-package :http)

;;; Steps:
;;; 1. Determine type from *http-get-types*
;;; 2. Follow application trees
;;; 3. Call function passing path & keyword arguments

;;; Type-based dispatch for GET, POST, HEAD and PUT methods

(defvar *lispweb-type-table* 
    '(("/robots.txt" . :robots)
      ("/image"      . :images)
      ("/class"      . :class)
      ("/map"        . :map)
      ("/mgd"        . :mgd)
      ("/gd"         . :gd)
      (""            . :text)))

(defgeneric http-deliver (type request)
  (:documentation "Output an object of the specified type."))

(defmethod http-deliver (type request)
  "Default method, look into application tree."
  (declare (ignore type))
  (or 
   (call-handler *user-pages* request)
   (call-handler *system-pages* request)
   (when *LispWeb-allow-html-files*
     (deliver-html-page request))
   (page-not-found request)))

(defun deliver-html-page (request &aux line)
  (let ((directory (pathname-directory (parse-namestring request))))
    (unless (or (member :up directory)
		(member :back directory))
      (with-open-file (input (lispweb-path request) 
			     :if-does-not-exist nil)
	(when input
	  (http-header *http-text-header* 
		       :code-name :ok 
		       :length (file-length input)
                       )
	  (loop 
	   (unless (setq line (read-line input nil nil))
	     (return))
	   (princ line *http-stream*)
	   (terpri *http-stream*))
	  (force-output *http-stream*)
          t)))))

(defmethod http-deliver ((type (eql :robots)) request)
  (declare (ignore request))
  (http-header *http-generic-header* :type "text/plain" :length (length *LispWeb-robots-txt*))
  (princ *LispWeb-robots-txt* *http-stream*))

#|
(defmethod http-deliver ((type (eql :images)) request)
  (let* ((filename (lispweb-path request))
	 (len (with-open-file (f filename) ;check for file presence
		(file-length f)))	;and get length
	 image)
    (http-header *http-generic-header* :type "image/gif" :length len)
    (setq image (gd-image-create-from-gif filename))
    (http-send-image image)
    (gd-image-destroy image)))
|#

(defmethod http-deliver ((type (eql :images)) request)
  (let ((filename (lispweb-path request)))
    (include-binary-file filename "image/gif")
    (http-force-output *http-stream*)))

(defmethod http-deliver ((type (eql :class)) request)
  (let ((filename (lispweb-path request)))
    (if (and (not (string= request ""))
	     (probe-file filename))
	(progn
	  (include-binary-file filename "application/octet-stream")
	  (http-force-output *http-stream*))
      (page-not-found request))))

(defmethod http-deliver ((type (eql :map)) request)
  (let (map x y mark1 mark2 args)
    (setq mark1 (position #\? request :from-end t))
    (setq mark2 (position #\, request))
    (cond ((and mark1 mark2)
	   (setq map (subseq request 0 mark1))
	   (setq x (read-from-string (subseq request (1+ mark1) mark2)))
	   (setq y (read-from-string (subseq request (1+ mark2))))
	   (setq args (list :map-x x :map-y y))
	   (or 
	    (call-handler *user-pages* map args)
	    (call-handler *system-pages* map args)
	    (page-not-found map)))
	  (t (bad-syntax-page request)))))

(defmethod http-deliver ((type (eql :netscape)) request)
  (declare (ignore request))
  :netscape)

(defmethod http-deliver ((type (eql :lynx)) request)
  (declare (ignore request))
  :lynx)

(defmethod http-deliver ((type (eql :generic)) request)
  (declare (ignore request))
  :generic)

(defmethod http-deliver ((type (eql :mosaic)) request)
  (declare (ignore request))
  :mosaic)

;;; Dispatch for PUT method

(defvar *lispweb-put-table*		;aggiungere /gif
    '(("/lisp" . :lisp)
      ("/file" . :file)))

#+ALLEGRO
(defmethod http-deliver ((type (eql :lisp)) request)
  (declare (ignore request))
  (let ((password (or (header-value :password) ""))
	(terminator (or (header-value :terminator) "end-of-put"))
	(flag nil))
    (if (check-password password)
	(progn
					;(http-header *http-text-header* :code-name :continue)
	  (let ((tempname (scm::make-temp-file-name))
		(compname "/no/such/file")
		(output nil)
		line)
	    (unwind-protect
		(progn
		  (with-open-file (file tempname :direction :output)
		    (loop
		      (setq line (http-read-line *http-stream* terminator))
		      (when (string= line terminator)
			(return))
		      (format file "~a~%" line)))

		  (setq output (with-output-to-string (out) 
				 (declare (special out))
				 (let ((*standard-output* (make-synonym-stream 'out))
				       (*error-output* (make-synonym-stream 'out))
				       (excl:*record-source-file-info* nil))
				   (setq compname (compile-file tempname)))))
		  (load compname)
		  (http-header *http-text-header* :code-name :ok :size (length output))
		  (princ output *http-stream*)
		  (setq flag t))
	      (progn
		(when (probe-file tempname)
		  (delete-file tempname))
		(when (probe-file compname)
		  (delete-file compname))
		(unless flag 
		  (http-header *http-text-header* :code-name :bad-syntax))))))
      (http-header *http-text-header* :code-name :no-auth))))

(defmethod http-deliver ((type (eql :file)) request)
  (let ((password (or (header-value :password) ""))
	(terminator (or (header-value :terminator) "end-of-put"))
	(line nil))
    (if (check-password password)
	(progn
	  (with-open-file (file (lispweb-path request) :direction :output)
	    (loop
	      (setq line (http-read-line *http-stream* terminator))
	      (when (string= line terminator)
		(return))
	      (format file "~a~%" line)))
	  (http-header *http-text-header* :code-name :ok))
      (http-header *http-text-header* :code-name :no-auth))))
