;;; -*- Mode: Common-Lisp; Package: :http -*-
;;; Program: LispWeb

;;; File: key.cl
;;; Description: 

;;; Author: Alberto Riva (alb@ipvaim.unipv.it)
;;; Created: Tue Apr  8, 1997
;;; Updated: Fri Apr 11, 1997, by Alberto Riva (alb@aim)
;;; Contents:
;;;   Top-level     IN-PACKAGE 
;;;   Top-level     ADD-LISPWEB-SERVICE 
;;;   Variable      *ACCESS-KEYS* 
;;;   Variable      *FORCE-ACCESS-KEY* 
;;;   Variable      *KEYED-CLASSES* 
;;;   Constant      *DEFAULT-KEY-DURATION* 
;;;   Constant      *DEFAULT-KEY-LENGTH* 
;;;   Constant      *ILLEGAL-KEY-CHARS* 
;;;   Structure     ACCESS-KEY 
;;;   Function      FIND-KEY 
;;;   Function      NEW-KEY 
;;;   Function      GENERATE-KEY-STRING 
;;;   Function      CHECK-KEY 
;;;   Function      KEY-EXPIRED-P 
;;;   Function      UPDATE-KEY 
;;;   Function      KEYED-CLASS-P 
;;; -*- -*-

(in-package :http)

(add-lispweb-service :key
		     :description "Expiring random key generator.")

(defvar *access-keys* nil)

(defvar *force-access-key* nil)

(defvar *keyed-classes* nil
  "Classes whose pages are protected by a key.")

(defconstant *default-key-duration*
    (time-to-seconds (make-time :minutes 15)))

(defconstant *default-key-length* 8)

(defconstant *illegal-key-chars* '(#\\ #\/ #\# #\: #\; #\& #\~ #\' #\Space #\")
  "Characters that cannot appear in a key.")

(defstruct access-key
  string
  time
  duration
  host)

(defun find-key (string)
  (find string *access-keys* 
	:key #'access-key-string
	:test #'string=))
		 
(defun new-key (host &optional (duration *default-key-duration*))
  (let ((key (make-access-key :string (generate-key-string)
			      :time (time-to-seconds (current-time))
			      :host host
			      :duration duration)))
    (push key *access-keys*)
    key))

(defun generate-key-string (&optional (length *default-key-length*))
  (labels ((random-char ()
	     (let ((char (code-char (+ 32 (random 94)))))
	       (if (member char *illegal-key-chars*)
		   (random-char)
		 char))))
    (let ((string (make-string 8)))
      (dotimes (i length)
	(setf (aref string i) (random-char)))
      string)))

(defun check-key (string)
  (let ((key (find-key string)))
    (cond (*force-access-key*
	   (setq key (find-key *force-access-key*))
	   (setq *force-access-key* nil)
	   (update-key key))
	  ((null key) nil)
	  ((key-expired-p key)
	   (setq *access-keys* (remove key *access-keys*))
	   :expired)
	  (t (update-key key)))))

(defun key-expired-p (key)
  (> (time-to-seconds (current-time)) 
     (+ (access-key-time key)
	(access-key-duration key))))

(defun update-key (key)
  (let ((new-key (generate-key-string)))
    (setf (access-key-string key) new-key)
    (setf (access-key-time key) (time-to-seconds (current-time)))
    new-key))

(defun keyed-class-p (class)
  "Determine whether a class needs keys."
  (member class *keyed-classes*))
