(in-package :http)

(defconstant *time-limits*
    (list most-positive-fixnum 12 31 24 60 60))

(defconstant *time-multipliers*
    (list (* 12 31 24 60 60)
	  (* 31 24 60 60)
	  (* 24 60 60)
	  (* 60 60)
	  60
	  1))

(defun make-time (&key (year 0)
		       (month 0)
		       (day 0)
		       (hour 0)
		       (minutes 0)
		       (seconds 0))
  "Create a TIME structure, with the specified
YEAR, MONTH, DAY, HOUR, MINUTES and SECONDS."
  (normalize-time (list year month day hour minutes seconds)))

(defun print-time (time &optional date-only)
  "Print TIME to string in 'DD-MM-YY HH:MM:SS' format; if
:date-only is T, just print the date."
  (if date-only
      (apply #'format nil "~2,'0d-~2,'0d-~d" 
	     (third time)
	     (second time)
	     (car time))
    (apply #'format nil "~2,'0d-~2,'0d-~d ~2,'0d:~2,'0d:~2,'0d" 
	   (third time)
	   (second time)
	   (car time)
	   (cdddr time))))

(defun parse-time (time-string)
  "Convert a TIME-STRING in the 'DD-MM-YY HH:MM:SS' or
'DD-MM-YY' format into a time structure."
  (let ((sl1 (position #\- time-string))
	(sp (position #\Space time-string))
	(col1 (position #\: time-string)))
    (let ((sl2 (when sl1 (position #\- time-string :start (1+ sl1))))
	  (col2 (when col1 (position #\: time-string :start (1+ col1)))))
      (if sp
	  (let ((day (safe-readnum (subseq time-string 0 sl1)))
		(month (safe-readnum (subseq time-string (1+ sl1) sl2)))
		(year (if sl2 
			  (safe-readnum (subseq time-string (1+ sl2) sp))
			0))
		(hour (safe-readnum (subseq time-string sp col1)))
		(minutes (safe-readnum (subseq time-string (1+ col1) col2)))
		(seconds (if col2
			     (safe-readnum (subseq time-string (1+ col2)))
			   0)))
	    (list year month day hour minutes seconds))
	(if sl1
	    (let ((day (safe-readnum (subseq time-string 0 sl1)))
		  (month (safe-readnum (subseq time-string (1+ sl1) sl2)))
		  (year (if sl2 
			    (safe-readnum (subseq time-string (1+ sl2)))
			  0))
		  (hour 0)
		  (minutes 0)
		  (seconds 0))
	      (list year month day hour minutes seconds))
	  (let ((day 0)
		(month 0)
		(year 0)
		(hour (safe-readnum (subseq time-string 0 col1)))
		(minutes (safe-readnum (subseq time-string (1+ col1) col2)))
		(seconds (if col2
			     (safe-readnum (subseq time-string (1+ col2)))
			   0)))
	    (list year month day hour minutes seconds)))))))

(defun add-times (time1 time2)
  "Add TIME1 and TIME2."
  (normalize-time (mapcar #'+ time1 time2)))

(defun subtract-times (time1 time2)
  "Subtract TIME2 from TIME2."
  (normalize-time (mapcar #'- time1 time2)))

(defun normalize-time (time &aux (carry 0))
  (declare (special carry))
  (normalize-time-aux time *time-limits*))

(defun normalize-time-aux (time limits)
  (declare (special carry))
  (let ((limit (car limits)))
    (cond ((null time) nil)
	  (t (let ((tail (normalize-time-aux (cdr time)
					     (cdr limits)))
		   (this (+ (car time) carry)))
	       (if (> this limit)
		   (multiple-value-setq (carry this)
		     (floor this limit))
		 (setq carry 0))
	       (cons this tail))))))

(defun time-to-seconds (time)
  "Convert TIME into the number of seconds since midnight, 
January 1st of the year 0."
  (reduce #'+ (mapcar #'* time *time-multipliers*)))

(defun time-to-days (time)
  "Convert TIME into the number of days since midnight, 
January 1st of the year 0."
  (/ (reduce #'+ (mapcar #'* 
			 (subseq time 0 3)
			 (subseq *time-multipliers* 0 3)))
     (third *time-multipliers*)))

(defun current-time ()
  "Return a time structure corresponding to the current time."
  (multiple-value-bind
      (second minute hour date month year day-of-week dst-p tz)
      (get-decoded-time)
    (declare (ignore tz day-of-week))
    (setq hour (+ hour (if dst-p 1 0)))
    (make-time :year year
	       :month month
	       :day date
	       :hour hour
	       :minutes minute
	       :seconds second)))

(defun time= (time1 time2)
  (when (stringp time1)
    (setq time1 (parse-time time1)))
  (when (stringp time2)
    (setq time2 (parse-time time2)))
  (= (time-to-seconds time1)
     (time-to-seconds time2)))

(defun time> (time1 time2)
  (when (stringp time1)
    (setq time1 (parse-time time1)))
  (when (stringp time2)
    (setq time2 (parse-time time2)))
  (> (time-to-seconds time1)
     (time-to-seconds time2)))

(defun time< (time1 time2)
  (when (stringp time1)
    (setq time1 (parse-time time1)))
  (when (stringp time2)
    (setq time2 (parse-time time2)))
  (< (time-to-seconds time1)
     (time-to-seconds time2)))

(defun time>= (time1 time2)
  (when (stringp time1)
    (setq time1 (parse-time time1)))
  (when (stringp time2)
    (setq time2 (parse-time time2)))
  (>= (time-to-seconds time1)
      (time-to-seconds time2)))

(defun time<= (time1 time2)
  (when (stringp time1)
    (setq time1 (parse-time time1)))
  (when (stringp time2)
    (setq time2 (parse-time time2)))
  (<= (time-to-seconds time1)
      (time-to-seconds time2)))

(defun time/= (time1 time2)
  (when (stringp time1)
    (setq time1 (parse-time time1)))
  (when (stringp time2)
    (setq time2 (parse-time time2)))
  (/= (time-to-seconds time1)
      (time-to-seconds time2)))
