;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

(in-package "OCML")


(in-ontology design-modification-ops)



(def-class DESIGN-MODIFICATION-OPERATOR (design-operator)
  ((applicable-to-parameters 
    :default-value '(setofall ?x (parameter ?x))
    :type function-expression
    :documentation "An expression which returns the set
                    of parameters whose value can be computed 
                    by means of this operator"))
  :lisp-class-name design-modification-operator )



(def-class BASIC-DESIGN-MODIFICATION-OPERATOR (design-modification-operator basic-operator) 
  "The body of a basic design modification operator is a 
   lambda expression which takes three arguments, a parameter,
   a value and a design model.
   The output is a list of assignments and deletions.  
   This operator can only spawn one successor state"
  ((has-body :type basic-design-modification-operator-body))
  :lisp-class-name basic-design-modification-operator )

(def-class BASIC-DESIGN-MODIFICATION-OPERATOR-BODY 
  (lambda-expression) ?x
  :constraint 
  (and (nth-domain ?x 1 parameter)
       (nth-domain ?x 3 design-model)
       (=> (= ?z (call ?p ?v ?d))
           (and  (has-value ?p ?v ?d)
                 (== ?z (?assignments ?deletions))
                 (set ?assignments)
                 (set ?deletions)
                 (every (append ?assignments ?deletions)
                        parameter-assignment)
                 (every ?deletions
                        (kappa (?pair)
                               (element-of ?pair ?d))))))
  :no-proofs-by (:constraint))

(def-class MULTIPLE-DESIGN-MODIFICATION-OPERATOR (design-modification-operator multiple-operator) 
  "The body of a multiple design modification operator is a 
   lambda expression which takes four arguments, a parameter,
   a value  a design model and a list of 'bad values' - i.e. values
   which have already been tried and failed
   The output is a list of assignments and deletions.  
   This operator can spawn multiple successor states"
  ((has-body :type multiple-design-modification-operator-body))
  :lisp-class-name multiple-design-modification-operator)

(def-class multiple-design-modification-operator-BODY (lambda-expression) ?x
  :constraint (and (nth-domain ?x 1 parameter)
                          (nth-domain ?x 3 design-model)
                          (=> (= ?z (call ?x ?p ?v ?d ?values))
                              (and 
                               (not (member ?z ?values))
                               (or (not (has-value-range ?p ?range))
                                   (and (has-value-range ?p ?range)
                                        (member ?z ?range)
                                        (forall ?v (=> (member ?v ?values)
                                                       (member ?v ?range)))))
                               (has-value ?p ?v ?d)
                               (== ?z (?assignments ?deletions))
                               (set ?assignments)
                               (set ?deletions)
                               (every (append ?assignments ?deletions)
                                      parameter-assignment)
                               (every ?deletions
                                      (kappa (?pair)
                                             (element-of ?pair ?d))))))
   :no-proofs-by (:constraint))

(defmethod  APPLY-DESIGN-OPERATOR ((design-op basic-design-modification-operator)
                                   (state design-state)  
                                   (focus parameter)
                                   context)
  context ;;;ignore
  (with-slots (applicability-env) design-op
    (ocml-apply (the-slot-value (name design-op) 'has-body)
                (List (name focus)
                      (right-value (name focus)(design-model state))
                      (list 'quote (design-model state)))
                applicability-env)))

(defmethod  APPLY-DESIGN-OPERATOR ((design-op basic-design-modification-operator)
                                   (state design-state)  
                                   focus
                                   context)
  context ;;;ignore
  (with-slots (applicability-env) design-op
    (ocml-apply (the-slot-value (name design-op) 'has-body)
                (List (List 'quote focus)
                      :nothing
                        (list 'quote (design-model state)))
                applicability-env)))

(defmethod  APPLY-DESIGN-OPERATOR ((design-op multiple-design-modification-operator)
                                   (state design-state)  
                                   focus
                                   context)
  context ;;;ignore
  (with-slots (applicability-env) design-op
    (ocml-apply (the-slot-value (name design-op) 'has-body)
                (List (name focus)
                      (right-value (name focus)(design-model state))
                       (list 'quote (design-model state))
                      (list 'quote (nogood-values state)))
                applicability-env)))


(defmethod TRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op multiple-design-modification-operator)
                                     (state design-state)  
                                     focus
                                     context)
  (let ((extensions (apply-design-operator 
                     design-op state  focus context)))
    (cond ((eq extensions :nothing)
           (backtrack-from-operator appl method state design-op focus context))
          (t
           (let ((deletions (compute-model-deletions 
                             appl method state design-op 
                             focus context extensions)))           
             (setf (nogood-values state)         ;;;This bit is not really needed for 
                   (cons (right-value (name focus) ;;;single-shot operators
                                      extensions) 
                         (nogood-values state)))
             (announce-state-changes extensions deletions)
             (add-design-state appl  method state  extensions deletions))))))

(defmethod TRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op basic-design-modification-operator)
                                     (state design-state)  
                                     focus
                                     context)
  (let ((extensions (apply-design-operator 
                     design-op state  focus context)))
    (cond ((eq extensions :nothing)
           (backtrack-from-operator appl method state design-op focus context))
          (t
           (let ((deletions (compute-model-deletions 
                             appl method state design-op 
                             focus context extensions)))           
             (announce-state-changes extensions deletions)
             (add-design-state appl  method state  extensions deletions))))))


(defmethod COMPUTE-MODEL-DELETIONS (appl (method design-method)
                                         (state design-state)  
                                         (design-op design-modification-operator)
                                         focus
                                         context
                                         extensions)
  "By default the deletions are the existing assignments of the 
   modified parameters" 
  appl focus context
  (mapcar #'(lambda (pair)
              (assoc (car pair) (design-model state)))
          extensions))


(defun announce-state-changes (extensions deletions)
  (when *trace-design-process?*
             (loop for ext in extensions
                   do
                   (print-with-spaces 2 ;;(state-depth state)
                                      "New assignment is <~a ~A>"
                                      (car ext)(cdr ext)))
             (loop for d in deletions
                   do
                   (print-with-spaces 2 ;;(state-depth state)
                                      "Deleting <~a ~A>" 
                                      (car d)(cdr d)))))

(defmethod RETRY-DESIGN-OPERATOR (appl (method design-method)
                                       (design-op basic-design-modification-operator)
                                       (state design-state)  
                                       focus
                                       context)
 
  (when *trace-design-process?*
    (print-with-spaces 2 ;;(state-depth state)
                       "Retrying ~a on ~a" 
                       (name design-op)
                       (name focus)))
  (backtrack-from-operator appl method state design-op focus context))

(defmethod RETRY-DESIGN-OPERATOR (appl (method design-method)
                                       (design-op multiple-design-modification-operator)
                                       (state design-state)  
                                       focus
                                       context)
 
 (when *trace-design-process?*
    (print-with-spaces 2 ;;; (state-depth state)
                       "Retrying ~a on ~a" 
                       (name design-op)
                       (name focus)))
  (try-design-operator appl method design-op state focus context))