;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

(in-package "OCML")

(in-ontology gen-design)


(def-class DESIGN-METHOD (problem-solving-method)
  ((applicable-to-task-type :value parametric-design)
   (has-input-role :value has-design-operators)
   (has-design-operators :type design-operator))
  :lisp-slots ((design-state-type :initform 'design-state
                                  :accessor design-state-type
                                  :initarg :design-state-type))
  
               ;;(initial-design-state-type :initform 'initial-design-state
               ;;                           :accessor initial-design-state-type
               ;;                           :initarg :initial-design-state-type))
   :lisp-class-name design-method)


(def-class DESIGN-APPLICATION (application)
  :Lisp-class-name design-application)


;;;This re-definition allows us to get a symbol-level view over parameters
(def-class parameter () ?p
  
  "A parameter is something which can have a value"
  ((depends-on :type parameter 
               :documentation 
               "This expresses the dependencies between parameters.
                A parameter, p1, depends on p2 if p2 has to be assigned in
                order to compute p1")
   (affects :type parameter))

  :iff-def (exists ?task (and (parametric-design ?task)
                              (has-parameters ?task ?l)
                              (member ?p ?l)))
  :lisp-class-name parameter
  :lisp-slots ((relevant-operators :accessor relevant-operators
                                     :type design-operator 
                                     :initform nil)))



;;;DESIGN-OPERATOR
(def-class design-operator ()
  ((applicable-to-parameters :type function-expression
                             :documentation "An expression which returns the set
                                             of parameters whose value can be computed 
                                             by means of this operator")
   (assumption :default-value (true)
               :type relation-expression
               :documentation 
               "This slot can be used to specify a statement that is expected
                   to hold for the application domain where the operator is applied.  
                   The difference between assumptions and preconditions is that the former
                   are static and the latter are dynamic.  The truth value of a precondition
                   might change during the design process.  Assumptions are expected to remain
                   (un-)satisfied during the design process")
   (has-precondition :default-value (true)
                     :type relation-expression
                     :documentation "This is an expression which can be used to determine whether
                                     an operator is applicable in the current state.  This expression
                                     should not depend on a particular design parameter")
   (body :type lambda-expression))
  :lisp-class-name design-operator
  :Lisp-slots (
               (effective-parameter-list :type parameter 
                                         :initform nil
                                         :accessor effective-parameter-list)
               (applicability-env :initform nil :accessor applicability-env)))


;;;DESIGN-OPERATOR-ORDER 
(def-relation design-operator-order (?x ?C)
  "This relation can be used to specify an application specific ordering
   of operators")

(def-rule design-operator-transitivity
  ((design-operator-order ?x ?C)
   if
   (asserted (design-operator-order ?x ?c2))
   (<> ?c ?c2)
   (design-operator-order ?c2 ?c)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun All-computable-parameters (parameters design-model)
  (filter parameters #'(lambda (p) (computable? p design-model))
          ))

(defmethod computable? ((p parameter) design-model)
  (every #'(lambda (param)
             (bound? (find-instance param) design-model))
         (get-slot-values p 'depends-on)))
                       



;;;COMPUTE-EFFECTIVE-APPLICABILITY-SET
(defmethod  compute-effective-applicability-set ((op design-operator))
  (ocml-eval-fun (the-slot-value (name op) 'applicable-to-parameters)))

(defmethod  PRECONDITION-HOLDS? ((op design-operator))
  (holds? (the-slot-value (name op) 'has-precondition)))


(defmethod  OPERATOR-STILL-USEFUL? ((op design-operator) d)
  "An operator is still useful if some of the parameters which can 
   be assigned a value by it are still unbound"
  (some #'(lambda (p)
            (unbound? p d))
        (effective-parameter-list op)))


