;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML;   -*-

(in-package "OCML")


(defvar *design-procedures*)
(defvar *design-fixes*)

(defmethod TOP-LEVEL-DESIGN (appl (method propose&revise))
  (setf  *design-procedures* 
         (mapcar #'(lambda (x)
                     (find-instance x 'design-procedure))
                 (the-slot-value (name method) 'has-design-procedures))
         *design-fixes* 
         (mapcar #'(lambda (x)
                     (find-instance x 'design-fix))
                 (the-slot-value (name method) 'has-design-fixes)))
  
  (cache-design-specification appl method)
  (cache-param-to-ops-links)
  (cache-param-to-constraints-links)
  (cache-constraints-to-fixes-links)
  (construct-dependency-network)
  (initialise-design-space appl method)
  (top-level-design-loop appl method))

(defun cache-param-to-ops-links ()
  (loop for op in *design-procedures*
        for params = (compute-effective-applicability-set op)
        do
        (loop for param in params
              do
              (link-param-to-operator (find-instance param 'parameter)
                                      op))))

(defun cache-constraints-to-fixes-links ()
  (loop for op in *design-fixes*
        for constraints = (compute-effective-applicability-set op)
        do
        (loop for c in constraints
              do
              (link-constraint-to-fix (find-instance c 'constraint)
                                      op))))

(defmethod link-constraint-to-fix ((c constraint) (op design-fix))
    (with-slots (effective-constraint-list) op
      (pushnew c effective-constraint-list)
      (pushnew op  (relevant-operators c))))

(defmethod  compute-effective-applicability-set ((d design-fix))
  (the-slot-value (name d) 'applicable-to-constraints))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod DESIGN-FROM-STATE (appl (method propose&revise) (state design-state))
  (if (state-feasible? state)
    (if (violated-constraints state)
      (revise-design appl method state)
      (if (state-complete? state)
        (design-succeeds state)
        (extend-design appl method state)))
    (state-fails state)))


(defmethod revise-DESIGN (appl (method design-method) (state design-state))
  (cond ((new-state? state)
         (setf (state-foci state) (collect-fixable-constraints appl method state)
               (new-state? state) nil)
         (design-from-context appl method state :revise))
        (t
         (backtrack-from-operator appl method state (current-operator state)
                                  (current-focus state) :revise))))


(defmethod COLLECT-FIXABLE-CONSTRAINTS (appl (method design-method) (state design-state))
  appl
  (with-slots (violated-constraints) state
    (let ((instances (all-instances 'fixable-parametric-constraint)))
      (filter violated-constraints 
              #'(lambda (x)
                  (and (member x instances) 
                       (relevant-operators x)))))))

(defmethod  APPLY-DESIGN-OPERATOR ((design-op design-fix)
                                   (state design-state)  
                                   (focus fixable-parametric-constraint)
                                   (context (eql :revise)))
  context ;;;ignore
  (let ((ass (right-value focus (bad-assignments state))))
    (ocml-call (the-slot-value (name design-op) 'body)
               (car ass)
               (cdr ass))))

(defmethod TRY-DESIGN-OPERATOR (appl (method design-method)
                                     (design-op design-modification-operator)
                                     (state design-state)  
                                     focus
                                     (context (eql :revise)))
  (let ((result (apply-design-operator 
                design-op state  focus context)))
  (cond ((eq result :nothing)
         (backtrack-from-operator appl method state design-op focus context))
        (t
         (let ((extensions (first result))
               (deletions (second result)))
           (when *trace-design-process?*
             (loop for ext in extensions
                   do
                   (print-with-spaces (state-depth state)
                                      "New assignment is <~a ~A>" (car ext)(cdr ext)))
             (loop for d in deletions
                   do
                   (print-with-spaces (state-depth state)
                                      "Deleting <~a ~A>" (car d)(cdr d))))
           (add-design-state appl  method state  extensions deletions))))))


   