;;; -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Package: OCML; -*- 

(in-package "OCML")

(def-ontology simple-classification
  :includes classification-task)

(in-ontology simple-classification)


(def-class simple-classification-psm (problem-solving-method)
  "This methods simple iterates over all candidates, classifying
   the instance with respect to each candidate and then 
   returning the best explaining classes"
  ((has-body (lambda (?psm)
               (simple-classification-procedure ?psm))))
  :own-slots ((tackles-task-type  classification-task))
  :lisp-class-name simple-classification-psm)

(def-procedure simple-classification-procedure (?psm)
  :lisp-fun #'(lambda (psm)
                (simple-classify psm)))

;;;A simple classification method.  It has to return the 
;;;classes which best explain the data - i.e. the instance.
(defun simple-classify (method)
  (let ((instance  (role-value method 'candidate-instance))
        (candidates  (role-value method 'candidate-classes))
        (features (setofall '?x `(has-feature ,instance , ?x))))
    (when candidates
      (loop with scores = (cons (achieve-generic-subtask
                                 method 'match-class 
                                 'candidate-instance instance
                                 'has-features features
                                 'candidate-class (car candidates))
                                (List (car candidates)))
            for c in (cdr candidates)
            for score = (achieve-generic-subtask
                         method 'match-class 
                         'candidate-instance instance
                         'has-features features
                         'candidate-class c)
            do
            (if (holds? `(better-fit ,score ,(car scores)))
              (setf scores (cons score
                                 (list c)))
              (when 
            (if (holds? ` (equal-fit ,score ,(car scores)))
                (push c (cdr scores))))
            finally
            (return (cdr scores)))))))

(def-task match-class (goal-specification-task) ?task 
  ((has-input-role :value candidate-class
                   :value has-features
                   :value candidate-instance)
   (has-output-role :value has-score)
   (candidate-class :type class)
   (candidate-instance :type instance)
   (has-features :type list)
   (has-score :type classification-score)
   (has-goal-expression 
    :value (kappa (?task ?score)
                  (and (classification-score ?score)
                       (has-score ?task ?score))))))


(def-class basic-match-class-psm (problem-solving-method)
  ((has-body (lambda (?psm)
               (simple-match-class ?psm))))
  :own-slots ((tackles-task-type  match-class))
  :lisp-class-name basic-match-class-psm)

(def-procedure simple-match-class (?psm)
  :lisp-fun #'(lambda (psm)
                (simple-match-class  psm)))

;;;SIMPLE-MATCH-CLASS
;;;Here I use a simple model of class matching.  GIven a feature f1 in instance I 
;;;and a class C, there can be 3 cases

;;;Feature is UNEXPLAINED - f1 is in I but not in C
;;;Feature is EXPLAINED - f1 is in I and in C and their values are consistent
;;;Feature is INCONSISTENT - f1 is in I and in C but their values are inconsistent

;;;It is also possible that teh class C has a feature fc, which is not in in I. 
;;;In this case we say that the feature is MISSING - i.e.

;;;Feature is MISSING - f1 is in C but not in I

;;;Hence, the result of matching a class to an instance is a 4-place vector
;;;<inconsistent-features, unexplained-features, explained-features, missing-features>

(defun simple-match-class (method)
  (let ((instance  (role-value method  'candidate-instance))
        (class  (role-value method  'candidate-class))
        (instance-features  (role-value method  'has-features))
        (class-features (setofall '?x `(has-feature ,class , ?x)))
        (score-vector (list nil nil nil nil)))

    ;;initialize unexplained features (in I but not in C)
    (setf (second score-vector)     
          (set-difference instance-features 
                          class-features))
    (loop for feature in instance-features
          for result = (achieve-generic-subtask
                        method 'match-feature 
                        'candidate-instance instance
                        'candidate-feature feature
                        'candidate-class class)
          do
          (case result
            (:inconsistent (setf (first score-vector)
                                 (cons feature (first score-vector))))
            (:unexplained (setf (second score-vector)
                                 (cons feature (second score-vector))))
            (:explained (setf (third score-vector)
                                 (cons feature (third score-vector)))))
          finally
          (progn 
            (setf (fourth score-vector)
                (set-difference class-features instance-features))
            (return score-vector)))))


(def-task match-feature (goal-specification-task) ?task 
  ((has-input-role :value candidate-class
                   :value candidate-feature
                   :value candidate-instance)
   (has-output-role :value has-feature-score)
   (candidate-class :type class)
   (candidate-instance :type instance)
   (candidate-feature :type list)
   (has-feature-score :type feature-score)
   (has-goal-expression 
    :value (kappa (?task ?score)
                  (and (feature-score ?score)
                       (has-feature-score ?task ?score))))))

(def-class basic-match-feature-psm (problem-solving-method)
  ((has-body (lambda (?psm)
               (simple-match-class-feature ?psm))))
  :own-slots ((tackles-task-type  match-feature))
  :lisp-class-name basic-match-feature-psm )

(def-procedure simple-match-class-feature (?psm)
  :lisp-fun #'(lambda (psm)
                (simple-match-class-feature psm)))

;;;Here we assume that an instance can have only one value for the
;;;feature in question, while a class can have multiple ones with
;;;an implicit disjunction - i..e an instance has to match only one 
;;;of these.
(defun simple-match-class-feature (method)
  (let ((instance  (role-value method  'candidate-instance))
        (class  (role-value method  'candidate-class))
        (feature  (role-value method  'candidate-feature))
        (class-features (setofall '?x `(has-feature ,class , ?x)))
        (instance-value (the-slot-value instance feature)))
    (cond ((member feature class-features)
           (let ((class-values (all-class-slot-values class feature)))
             (if class-values
               (if (some #'(lambda (value)
                             (not (eq :fail
                                      (unify instance-value
                                             value))))
                       class-values)
                 :explained
                 :inconsistent)
               :unexplained)))
          (t
           :unexplained))))


  



  






