package org.infrawebs.axiom.editor;

import java.util.List;
import java.util.Observable;
import java.util.Set;

import org.eclipse.core.runtime.IPath;
import org.infrawebs.axiom.logexpr.WsmoLogExprExport;
import org.infrawebs.axiom.logexpr.WsmoLogExprImport;
import org.infrawebs.axiom.logexpr.WsmoTopEntityExport;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.operations.Persistence;
import org.infrawebs.common.Util;
import org.omwg.logicalexpression.LogicalExpression;
import org.omwg.ontology.Axiom;
import org.omwg.ontology.Ontology;
import org.wsmo.common.TopEntity;
import org.wsmo.service.WebService;
import org.wsmostudio.runtime.WSMORuntime;
import org.wsmostudio.ui.editors.model.AxiomModel;
import org.wsmostudio.ui.editors.model.ObservableModel;

public class AxiomEditorContents extends Observable
{
	private WsmoLogExprExport leExport = new WsmoLogExprExport();
	private WsmoTopEntityExport entityExport = new WsmoTopEntityExport();

	private AxiomContentsModel axiomContentsModel = null;
	private LogicalExpression logicalExpression = null;

	// filePath and studioModel are mutually exclusive. At least one of them must be null.
	private IPath filePath = null;
	private AxiomModel studioModel = null; 
	
	private Persistence persistence = new Persistence();

	public AxiomEditorContents() {
		super();
	}

	public AxiomContentsModel getAxiomContentsModel() {
		return this.axiomContentsModel;
	}

	public LogicalExpression getLogicalExpression() {
		return this.logicalExpression;
	}

	public IPath getFilePath() {
		return this.filePath;
	}

	public AxiomModel getStudioModel() {
		return this.studioModel;
	}

	public void initFromStudio(AxiomModel studioAxiomModel) {
		this.axiomContentsModel = null;
		this.logicalExpression = null;
		this.filePath = null;
		this.studioModel = null;

		if (studioAxiomModel == null || studioAxiomModel.getAxiom() == null) {
			return;
		}

		this.studioModel = studioAxiomModel;
		Axiom axiom = studioAxiomModel.getAxiom();
		String axiomName = null;
		try {
			axiomName = Util.getLocalName(axiom); 
		}
		catch (Exception ex) {
			axiomName = "noname";
		}

		Set definitions = axiom.listDefinitions();
		if (definitions.size() > 0) {
			// If at least one definition is present, use the first one.
			this.logicalExpression = (LogicalExpression)definitions.iterator().next();

			// Find the imported ontologies of the context top entity.
			List<Ontology> importedOntos = Util.listImportedOntos(getTopEntity());

			this.axiomContentsModel = WsmoLogExprImport.createAxiomModel(
					this.logicalExpression, axiomName, importedOntos);
		}
		else {
			// If no definitions are present, create an empty model and leave the LE to be null.
			// Thus when the model is changed a new definition will be added to the studio axiom.
			this.axiomContentsModel = new AxiomContentsModel(axiomName);
		}
	}

	public void initFromFile(IPath path) {
		this.axiomContentsModel = this.persistence.load(path);
		this.logicalExpression = null;
		this.filePath = path;
		this.studioModel = null;
		this.setChanged();
	}

	public void initEmpty(String axiomName) {
		this.axiomContentsModel = new AxiomContentsModel(axiomName);
		this.logicalExpression = null;
		this.filePath = null;
		this.studioModel = null;
		this.setChanged();
	}

	/**
	 * Prepares the new LogicalExpression for the observers to update themselves.
	 * For example, the AxiomTextView will regenerate its text.
	 */
	@Override
	public void setChanged() {
		super.setChanged();

		if (this.axiomContentsModel == null) {
			return;
		}

		if (this.studioModel != null && this.studioModel.getAxiom() != null) {
			// Replace the old definition in the model provided by WSMO Studio.
			// NOTE: If there is no old definition, a new one is added. As a result
			// new axioms are automatically initialized with a first definition.
			// NOTE: If the axiom model is empty, the old definition is deleted.
			// So, to delete a definition the user has to detach the Start element.
			if (this.logicalExpression != null) {
				this.studioModel.removeDefinition(this.logicalExpression);
			}
			this.logicalExpression = this.leExport.createLogExpr(this.axiomContentsModel);
			if (this.logicalExpression != null) {
				this.studioModel.addDefinition(this.logicalExpression);
			}
		}
		else {
			// We are not editing a WSMO Studio object. Just rebuild the expression.
			this.logicalExpression = this.leExport.createLogExpr(this.axiomContentsModel);
		}

		notifyObservers();
	}

	public boolean saveToFile(IPath newFilePath) {
		boolean success = this.persistence.save(
				this.leExport.listUsedOntologies(),
				this.axiomContentsModel,
				newFilePath.toString());
		if (success) {
			this.filePath = newFilePath;
			// Detach from WSMO Studio. From now on the editor will save to the file system only.
			this.studioModel = null;
		}
		return success;
	}

	private TopEntity getTopEntity() {
		if (this.studioModel == null) {
			return null;
		}
        ObservableModel model = this.studioModel;
        while (model.getMasterModel() != null) {
            model = model.getMasterModel();
        }
        return (TopEntity)model.getAdapter(TopEntity.class);
	}

	public void saveToStudio() throws Exception {
		if (this.studioModel != null) {
			// Save the top entity in WSMO Studio.
			WSMORuntime.getRuntime().doUpdateEntity(getTopEntity());
		}
	}

	public WebService createEnclosingWebService() {
		if (this.axiomContentsModel == null) {
			return null;
		}

		Axiom axiom = this.entityExport.createAxiom(
				this.logicalExpression,
				this.axiomContentsModel.getAxiomRoot().getAxiomName(),
				this.axiomContentsModel.getAxiomRoot().getNFPs());
		return this.entityExport.createWebService(
				this.leExport.listUsedOntologies(), axiom);
	}

}
