package org.infrawebs.axiom.logexpr;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.infrawebs.axiom.model.AttributeModel;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.model.InstanceModel;
import org.infrawebs.axiom.model.MemberOfModel;
import org.infrawebs.axiom.model.ModelFactory;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.OperatorModel;
import org.infrawebs.axiom.model.ParameterModel;
import org.infrawebs.axiom.model.RelationModel;
import org.infrawebs.axiom.model.ValueInstanceModel;
import org.infrawebs.axiom.model.VariableModel;
import org.infrawebs.axiom.operations.ModelOperations;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.RestorationException;
import org.omwg.logicalexpression.Atom;
import org.omwg.logicalexpression.AttributeValueMolecule;
import org.omwg.logicalexpression.MembershipMolecule;
import org.omwg.logicalexpression.terms.Term;
import org.omwg.ontology.Concept;
import org.omwg.ontology.DataValue;
import org.omwg.ontology.Instance;
import org.omwg.ontology.Ontology;
import org.omwg.ontology.OntologyElement;
import org.omwg.ontology.Relation;
import org.omwg.ontology.Variable;
import org.wsmo.common.Identifier;
import org.wsmo.common.UnnumberedAnonymousID;
import org.wsmostudio.runtime.WSMORuntime;

/**
 * A combination of memberships, refinements, relations and operators.
 * All of them are just conjunction operands in WSMO logical expressions.
 * In the Axiom Model however they form complex structures.
 */
public class ComplexConjunction {
	// MemberOfModels, RelationModels or OperatorModels
	private Set<ModelWithConnections> models = new HashSet<ModelWithConnections>();
	// key: variable name; value: MemberOfModel or OperatorModel
	private Map<String, ModelWithConnections> varNameOwners = new HashMap<String, ModelWithConnections>();
	private Set<RelationModel> relations = new HashSet<RelationModel>();
	private Map<OperatorModel, ComplexOperator> operators = new HashMap<OperatorModel, ComplexOperator>();
	private Set<AttributeValueMolecule> attributeValues = new HashSet<AttributeValueMolecule>();
	private Map<Variable, DataValue> assignments = new HashMap<Variable, DataValue>();

	private AxiomContentsModel contents = null;
	private ModelFactory factory = null;
	private List<Ontology> importedOntos = null;

	public ComplexConjunction(AxiomContentsModel contents, List<Ontology> importedOntos) {
		this.contents = contents;
		this.factory = contents.getFactory();
		this.importedOntos = importedOntos;
	}

	private void internalAddMemberOf(MemberOfModel model, Variable varTerm) {
		// Replace the default generated VarNameModel with the one used in the expression. 
		model.setVarNameModel(WsmoLogExprImport.getVarNameModel(varTerm, this.contents));
		this.models.add(model);
		this.varNameOwners.put(varTerm.getName(), model);
		this.contents.addChild(model);
	}

	/**
	 * NOTE: The expression must be the equality predicate. 
	 */
	public void addInstance(Atom expr) throws RestorationException {
		// Validate the atom's structure.
		if (expr.getArity() != 2 || !(expr.getParameter(0) instanceof Variable)) {
			return;
		}
		internalAddInstance((Variable)expr.getParameter(0), expr.getParameter(1));
	}

	/**
	 * Depending on the type of the second parameter, either creates an instance immediately
	 * or remembers that a value-instance must be created at "compile time".
	 */
	private void internalAddInstance(Variable varTerm, Term valueTerm) throws RestorationException {
		if (!(valueTerm instanceof Identifier) && !(valueTerm instanceof DataValue)) {
			return;
		}
		
		if (valueTerm instanceof Identifier) {
			OntologyElement e = AxiomEditorPlugin.ontoLoader.findDefinition(
					(Identifier)valueTerm, Instance.class, this.importedOntos);
			InstanceModel instance = (InstanceModel)this.factory.createAxiomElement(e);
			internalAddMemberOf(instance, varTerm);
		}
		else {  // valueTerm instanceof DataValue
			// Postpone the creation of the value-instance until its membership is known.
			this.assignments.put(varTerm, (DataValue)valueTerm);
		}
	}

	/**
	 * NOTE: The expression should NOT be the equality predicate. 
	 */
	public void addRelation(Atom expr) throws RestorationException {
		Identifier id = expr.getIdentifier();
		OntologyElement e = AxiomEditorPlugin.ontoLoader.findDefinition(
				id, Relation.class, this.importedOntos);
		RelationModel relation = (RelationModel)this.factory.createAxiomElement(e);
		List params = relation.getRefinableList().getChildren();
		if (relation == null || expr.getArity() != params.size()) {
			return;
		}

		// Bind the relation model's parameters to the variable terms.
		for (int i = 0; i < expr.getArity(); i++) {
			Term valueTerm = expr.getParameter(i); 
			ParameterModel param = (ParameterModel)params.get(i);
			if (valueTerm instanceof Variable) {
				param.setVarNameModel(WsmoLogExprImport.getVarNameModel((Variable)valueTerm, this.contents));
			}
			else if (valueTerm instanceof UnnumberedAnonymousID) {
				// Skip not bound parameters.
			}
			// Invent an intermediate variable name for the direct parameter assignments. 
			else if (valueTerm instanceof Identifier || valueTerm instanceof DataValue) {
				// KNOWN BUG: There might be a naming conflict because not all variables
				// have been craeted by this moment.
				String varName = Util.getLocalName(param.getRefinable());
				Variable varTerm = WSMORuntime.getRuntime().getWsmoFactory().createVariable(varName);
				internalAddInstance(varTerm, valueTerm);

				param.setVarNameModel(WsmoLogExprImport.getVarNameModel(varTerm, this.contents));
			}
			else {
				// Invalid parameter term.
			}
		}

		this.models.add(relation);
		this.relations.add(relation);
		this.contents.addChild(relation);
	}

	public void addMembership(MembershipMolecule expr) throws RestorationException {
		// Validate the membership's structure.
		if (!(expr.getLeftParameter() instanceof Variable) ||
				!(expr.getRightParameter() instanceof Identifier)) {
			return;
		}

		Identifier typeTerm = (Identifier)expr.getRightParameter(); 
		Concept concept = (Concept)AxiomEditorPlugin.ontoLoader.findDefinition(
				typeTerm, Concept.class, this.importedOntos);

		// Depending on the type, this will create a variable or a value-instance.
		MemberOfModel memberOf = (MemberOfModel)this.factory.createAxiomElement(concept);

		Variable varTerm = (Variable)expr.getLeftParameter();
		internalAddMemberOf(memberOf, varTerm);
	}

	/**
	 * Queues assigments of attributes to variables because they will be bound to the attribute
	 * models at "compile time".
	 * Creates intermediate instances or value-instances if the attribute is directly assigned a value.
	 */ 
	public void addAttributeValue(AttributeValueMolecule expr) throws RestorationException {
		// Validate the refinements' structure.
		if (!(expr.getAttribute() instanceof Identifier) ||
				!(expr.getLeftParameter() instanceof Variable)) {
			return;
		}

		// Invent an intermediate variable name for the direct attribute assignments. 
		Term valueTerm = expr.getRightParameter();
		if (valueTerm instanceof Identifier || valueTerm instanceof DataValue) {
			// KNOWN BUG: There might be a naming conflict because not all variables
			// have been craeted by this moment.
			String varName = Util.getLocalName(expr.getAttribute());
			Variable varTerm = WSMORuntime.getRuntime().getWsmoFactory().createVariable(varName);
			internalAddInstance(varTerm, valueTerm);

			// Replace "?x[attr hasValue 3.14]" with "?x[attr hasValue ?attr1]" and queue it.
			this.attributeValues.add(WsmoLogExprExport.leFactory.createAttributeValue(
					expr.getLeftParameter(), expr.getAttribute(), varTerm));
		}
		// Queue an expression of the form ?x[attr hasValue ?y].
		else if (valueTerm instanceof Variable) {
			this.attributeValues.add(expr);
		}
	}

	public void addOperator(ComplexOperator operator) {
		this.models.add(operator.getOperator());
		this.operators.put(operator.getOperator(), operator);
		this.contents.addChild(operator.getOperator());
	}

	/**
	 * Triggers the recursive compilation of all subexpressions.
	 * Associates the operators with the variables they represent (if they represent any). 
	 */
	private void compileOperators() {
		for (ComplexOperator operator: this.operators.values()) {
			operator.compile();
			String varName = operator.getPrimaryVariableName();
			if (varName != null) {
				this.varNameOwners.put(varName, operator.getOperator());
			}
		}
	}

	/**
	 * Sets the values of all value-instances.
	 * Creates those which have not been created from a membership.
	 */
	private void bindAssignments() {
		for (Entry<Variable, DataValue> entry: this.assignments.entrySet()) {
			// Find the value-instance model whose value must be assigned.
			Variable varTerm = entry.getKey();
			DataValue data = entry.getValue();
			Object varNameOwner = this.varNameOwners.get(varTerm.getName());
			
			ValueInstanceModel valueInstance = null;
			if (varNameOwner instanceof ValueInstanceModel) {
				valueInstance = (ValueInstanceModel)varNameOwner;
			}
			else {
				// Create the value-instance with a default type if its type
				// was not explicitely stated with a membership. 
				Concept concept = AxiomEditorPlugin.ontoStore.liftTypeToConcept(data.getType()); 
				valueInstance = (ValueInstanceModel)this.factory.createAxiomElement(concept);
				internalAddMemberOf(valueInstance, varTerm);
			}

			valueInstance.setValue(data.toString());
		}
		this.assignments.clear();
	}

	/**
	 * Finds a free attruibute with the same identifier as in the expression,
	 * binds it to a VarNameModel and connects it to a refining model.
	 * Excludes the refining model from the list of models to simplify this
	 * conjunction. The models which remain last will be the only explicit
	 * operands of this complex conjunction.
	 */
	private void bindAttribute(VariableModel variable, AttributeValueMolecule expr) {
		for (ModelWithConnections child: variable.getRefinableList().getChildren()) {
			AttributeModel attribute = (AttributeModel)child;
			if (attribute.getRefinable().getIdentifier().equals(expr.getAttribute()) &&
					attribute.getVarNameModel() == null) {
				Variable varTerm = (Variable)expr.getRightParameter();
				attribute.setVarNameModel(WsmoLogExprImport.getVarNameModel(varTerm, this.contents));
				ModelWithConnections refiningModel = this.varNameOwners.get(varTerm.getName());
				if (refiningModel != null) {
					ModelOperations.connect(attribute, refiningModel);
					if (this.models.size() > 1) {
						this.models.remove(refiningModel);
					}
				}
				break;
			}
		}
	}

	private void bindAttributes() {
		for (AttributeValueMolecule expr: this.attributeValues) {
			// Find the variable model whose attribute must be refined.
			Variable hostVarTerm = (Variable)expr.getLeftParameter();
			Object varNameOwner = this.varNameOwners.get(hostVarTerm.getName());
			if (varNameOwner instanceof VariableModel) {
				bindAttribute((VariableModel)varNameOwner, expr);
			}
			else {
				// Handle orphaned refinement...
			}
		}
		this.attributeValues.clear();
	}

	/**
	 * Connects the paremeters to refining elements and excludes *some* elements from
	 * the list of models to simplify this conjunction. The models which remain last
	 * will be the only explicit operands of this complex conjunction.
	 */
	private void bindParameters() {
		for (RelationModel relation: this.relations) {
			for (ModelWithConnections child: relation.getRefinableList().getChildren()) {
				ParameterModel parameter = (ParameterModel)child;
				if (parameter.getVarNameModel() != null) {
					ModelWithConnections refiningModel = this.varNameOwners.get(
							parameter.getVarNameModel().getName());
					if (refiningModel != null) {
						ModelOperations.connect(parameter, refiningModel);
						if (this.models.size() > 1) {
							// TODO: Exclude what?
							// First of all, exclude the relation (thus making it floating) and
							// leave the first refined parameter to be the root. Then remove the
							// refining models of all the other parameters.
							this.models.remove(this.models.contains(relation)
									? relation : refiningModel);
						}
					}
				}
			}
		}
		this.relations.clear();
	}

	/**
	 * @return A variable name if all models represent the same variable.
	 *         Null if any two models represent different variables.
	 */
	public String getPrimaryVariableName() {
		String result = null;
		for (ModelWithConnections model: this.models) {
			String varName = null;
			if (model instanceof RelationModel) {
				continue;  // Ignore relations (hope they are just floating relations).
			}
			else if (model instanceof MemberOfModel) {
				varName = ((MemberOfModel)model).getVarNameModel().getName();
			}
			else if (model instanceof OperatorModel) {
				varName = this.operators.get(model).getPrimaryVariableName();
			}
			
			if (varName == null) {
				return null;
			}
			else if (result == null) {
				result = varName;
			}
			else if (!result.equals(varName)) {
				return null;
			}
		}
		return result;
	}

	/**
	 * If there are several operands connects them with an explicit logical AND operator.
	 */
	private ModelWithConnections createTopModel() {
		if (this.models.size() >= 2) {
			OperatorModel operatorAnd = this.factory.createOperator(OperatorModel.OPERATOR_AND, null);
			this.contents.addChild(operatorAnd);
			for (ModelWithConnections model: this.models) {
				ModelOperations.connect(operatorAnd, model); 
			}
			return operatorAnd;
		}
		if (this.models.size() == 1) {
			return this.models.iterator().next();
		}
		return null;
	}
	
	/**
	 * Compiles all gathered elements into a single Axiom Model fragment.
	 * Must be called after all the operands of the conjunction have been added.
	 * @return The element which is the root of the compiled Axiom Model fragment. 
	 */
	public ModelWithConnections compile() {
		compileOperators();
		bindAssignments();
		bindAttributes();
		bindParameters();
		return createTopModel();
	}

}

