package org.infrawebs.axiom.logexpr;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import org.deri.wsmo4j.logicalexpression.AbstractVisitor;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.VarNameList;
import org.infrawebs.axiom.model.VarNameModel;
import org.infrawebs.axiom.operations.ModelOperations;
import org.infrawebs.common.Messages;
import org.infrawebs.ontology.RestorationException;
import org.omwg.logicalexpression.Atom;
import org.omwg.logicalexpression.AttributeConstraintMolecule;
import org.omwg.logicalexpression.AttributeInferenceMolecule;
import org.omwg.logicalexpression.AttributeValueMolecule;
import org.omwg.logicalexpression.CompoundMolecule;
import org.omwg.logicalexpression.Conjunction;
import org.omwg.logicalexpression.Constants;
import org.omwg.logicalexpression.Disjunction;
import org.omwg.logicalexpression.ExistentialQuantification;
import org.omwg.logicalexpression.LogicalExpression;
import org.omwg.logicalexpression.MembershipMolecule;
import org.omwg.logicalexpression.Negation;
import org.omwg.logicalexpression.SubConceptMolecule;
import org.omwg.ontology.Ontology;
import org.omwg.ontology.Variable;

public class WsmoLogExprImport
{
	/**
	 * Searches the axiom's list of variables for a VarNameModel with same name as the
	 * variable specified. Creates the model if it doesn't exist.
	 * NOTE: The new VarNameModel is not added to the list of variables until it is
	 * associated with at least one axiom element (VarNameHolderModel).
	 * @return The VarNameModel corresponding to the variable.
	 */
	public static VarNameModel getVarNameModel(Variable variable, AxiomContentsModel contents) {
		String varName = variable.getName();
		VarNameList list = contents.getVarNameList();
		VarNameModel result = list.get(varName);
		if (result == null) {
			result = new VarNameModel(list, varName);
		}
		return result;
	}

	/**
	 * Reconstructs the axiom model from a WSMO Logical expression.
	 */
	public static AxiomContentsModel createAxiomModel(LogicalExpression le,
			String axiomName, List<Ontology> importedOntos) {
		if (le == null) {
			return null;
		}

		AxiomContentsModel contents = new AxiomContentsModel(axiomName);
		AxiomModelBuilder amb = new AxiomModelBuilder(importedOntos);
		amb.compile(le, contents);
		return contents;
	}
}

class AxiomModelBuilder extends AbstractVisitor
{
	private AxiomContentsModel contents = null;
	private List<Ontology> importedOntos = null;

	private Stack<ComplexConjunction> conjStack = new Stack<ComplexConjunction>();
	// A stack of operators used to merge the operands of adjacent disjunctions.
	private Stack<ComplexOperator> operatorStack = new Stack<ComplexOperator>();
	private Set<String> existsVariables = new HashSet<String>(); 

	public AxiomModelBuilder(List<Ontology> importedOntos) {
		this.importedOntos = importedOntos;
	}

	private void pushNewConjunction() {
		this.conjStack.push(new ComplexConjunction(this.contents, this.importedOntos));
	}

	private ComplexConjunction popCompleteConjunction() {
		return this.conjStack.pop();
	}

	/**
	 * Appends the conjunction's operands to the current ComplexConjunction.
	 */
	@Override
	public void visitConjunction(Conjunction expr) {
		if (this.operatorStack.peek() instanceof DummyConjunction) {
			// Merge this conjunction with the already started conjunction.
			super.visitConjunction(expr);
		}
		else {
			this.operatorStack.push(DummyConjunction.instance);
			super.visitConjunction(expr);
			this.operatorStack.pop();
		}
	}

	/**
	 * Merges the operands of consecutive binary operators into a single N-ary operator.
	 */
	@Override
	public void visitDisjunction(Disjunction expr) {
		if (this.operatorStack.peek() instanceof ComplexDisjunction) {
			// Merge this disjunction with the already started disjunction.
			ComplexDisjunction disjunction = (ComplexDisjunction)this.operatorStack.peek(); 

			expr.getLeftOperand().accept(this);

			pushNewConjunction();
			expr.getRightOperand().accept(this);
			disjunction.addOperand(popCompleteConjunction());
		}
		else {
			ComplexDisjunction disjunction = new ComplexDisjunction(this.contents); 
			this.operatorStack.push(disjunction);
			
			pushNewConjunction();
			expr.getLeftOperand().accept(this);
			disjunction.addOperand(popCompleteConjunction());

			pushNewConjunction();
			expr.getRightOperand().accept(this);
			disjunction.addOperand(popCompleteConjunction());

			this.operatorStack.pop();
			this.conjStack.peek().addOperator(disjunction);
		}
	}

	@Override
	public void visitNegation(Negation expr) {
		ComplexNegation negation = new ComplexNegation(this.contents);
		this.operatorStack.push(negation);

		// Create the subtree rooted at the negation operand.
		pushNewConjunction();
		super.visitNegation(expr);
		negation.setOperand(popCompleteConjunction());

		this.operatorStack.pop();
		this.conjStack.peek().addOperator(negation);
	}

	/**
	 * An atom can be translated into an instance/value-instance or into a relation.
	 */
	@Override
	public void visitAtom(Atom expr) {
		try {
			if (expr.getIdentifier().toString().equals(Constants.EQUAL)) {
				this.conjStack.peek().addInstance(expr);
			}
			else {
				this.conjStack.peek().addRelation(expr);
			}
			super.visitAtom(expr);
		}
		catch (RestorationException ex) {
			Messages.showWarning(ex.getMessage());
		}
	}

	/**
	 * Visits all memberships and attribute refinements.
	 */
	public void visitCompoundMolecule(CompoundMolecule expr) {
		Iterator m = expr.listMemberShipMolecules().iterator();
		while (m.hasNext()) {
			MembershipMolecule membership = (MembershipMolecule)m.next();
			visitMemberShipMolecule(membership);
		}

		Iterator a = expr.listAttributeValueMolecules().iterator();
		while (a.hasNext()) {
			AttributeValueMolecule refinement = (AttributeValueMolecule)a.next();
			visitAttributeValueMolecule(refinement);
		}
	}

	public void visitMemberShipMolecule(MembershipMolecule expr) {
		try {
			this.conjStack.peek().addMembership(expr);
		}
		catch (RestorationException ex) {
			Messages.showWarning(ex.getMessage());
		}
	}

	public void visitAttributeValueMolecule(AttributeValueMolecule expr) {
		try {
			this.conjStack.peek().addAttributeValue(expr);
		}
		catch (RestorationException ex) {
			Messages.showError(ex.getMessage());
		}
	}

	@Override
    public void visitExistentialQuantification(ExistentialQuantification expr) {
    	// Just gather the names of the variables. The VarNameModels will be marked later.
    	Iterator i = expr.listVariables().iterator();
    	while (i.hasNext()) {
    		Variable varTerm = (Variable)i.next(); 
        	this.existsVariables.add(varTerm.getName());
    	}
    	super.visitExistentialQuantification(expr);
    }

	public void visitSubConceptMolecule(SubConceptMolecule expr) {
		// Obligation to implement method.
	}

	public void visitAttributeContraintMolecule(AttributeConstraintMolecule expr) {
		// Obligation to implement method.
	}

	public void visitAttributeInferenceMolecule(AttributeInferenceMolecule expr) {
		// Obligation to implement method.
	}

	/**
	 * Marks the VarNameModels which were declared in an existential quantifier.
	 * Their names have already been gathered in existsVariables. 
	 */
	private void setExistsFlags() {
    	for (String varName: this.existsVariables) {
        	VarNameModel model = this.contents.getVarNameList().get(varName);
        	if (model != null) {
        		model.setExistsFlag(true);
        	}
    	}
	}

    /**
	 * Reconstructs the model graph from a WSMO Logical expression.
	 */
	public void compile(LogicalExpression le, AxiomContentsModel axiomContents) {
		this.contents = axiomContents;

		// Create the top-level conjunction.
		this.conjStack.clear();
		pushNewConjunction();
		
		// Add a dummy operator to avoid EmptyStackExceptions when peeking.
		this.operatorStack.clear();
		this.operatorStack.push(null);
		
		this.existsVariables.clear();

		// Preprocess the logical expression: create a tree of models and ComplexConjunctions
		// and gather the existentially-quantified variables.
		le.accept(this);

		// There should be only one conjunction left in the stack (the top one).
		// Compiling it should trigger recursive compilation of everything else.
		ModelWithConnections topModel = this.conjStack.peek().compile();
		ModelOperations.connect(this.contents.getAxiomRoot(), topModel);

		setExistsFlags();
	}
}

