package org.infrawebs.axiom.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

public abstract class ModelWithConnections extends AbstractModel
{
	public static final int UNKNOWN_DEPTH = -1;
	public static final String ID_PROPERTY_SOURCE_CONNECTION = "ID_PROPERTY_SOURCE_CONNECTION";
	public static final String ID_PROPERTY_TARGET_CONNECTION = "ID_PROPERTY_TARGET_CONNECTION";

	private static final long serialVersionUID = 200512242100L;

	private List<ConnectionModel> sourceConnections = new ArrayList<ConnectionModel>();
	private List<ConnectionModel> targetConnections = new ArrayList<ConnectionModel>();
	protected AxiomContentsModel contents = null;

	public ModelWithConnections(AxiomContentsModel contents) {
		this.contents = contents; 
	}

	protected void addSourceConnection(ConnectionModel connx) {
		this.sourceConnections.add(connx);
		firePropertyChange(ID_PROPERTY_SOURCE_CONNECTION, null, null);
	}

	protected void addTargetConnection(ConnectionModel connx) {
		this.targetConnections.add(connx);
		firePropertyChange(ID_PROPERTY_TARGET_CONNECTION, null, null);
	}

	public List<ConnectionModel> listSourceConnections() {
		return this.sourceConnections;
	}

	public List<ConnectionModel> listTargetConnections() {
		return this.targetConnections;
	}

	public AxiomContentsModel getContents() {
		return this.contents;
	}

	protected void removeSourceConnection(Object connx) {
		this.sourceConnections.remove(connx);
		firePropertyChange(ID_PROPERTY_SOURCE_CONNECTION, null, null);
	}

	protected void removeTargetConnection(Object connx) {
		this.targetConnections.remove(connx);
		firePropertyChange(ID_PROPERTY_TARGET_CONNECTION, null, null);
	}

	// Lists the models at the remote ends of the connections outgoing from this model.
	public List<ModelWithConnections> listGrafChildren() {
		List<ModelWithConnections> children = new ArrayList<ModelWithConnections>();
		Iterator iter = listSourceConnections().iterator();
		while (iter.hasNext()) {
			ConnectionModel conn = (ConnectionModel)iter.next(); 
			ModelWithConnections child = conn.getTarget();
			if (child != null) {
				children.add(child);
			}
		}
		return children;
	}

	// Lists the models at the remote ends of the connections coming into this model.
	public List<ModelWithConnections> listGrafParents() {
		List<ModelWithConnections> parents = new ArrayList<ModelWithConnections>();
		Iterator iter = listTargetConnections().iterator();
		while (iter.hasNext()) {
			ConnectionModel conn = (ConnectionModel)iter.next(); 
			ModelWithConnections parent = conn.getSource();
			if (parent != null) {
				parents.add(parent);
			}
		}
		return parents;
	}

	/**
	 * @see MemberOfModel.listChildRelations()
	 * @return True if this model has a parent model which completely
	 * defines the model's position in the logical expression. False if the
	 * position is inferred from the alignment of the child models. 
	 */
	public boolean isFixedInExpression() {
		return (listTargetConnections().size() > 0);
	}
	
	/**
	 * Searches the connection chains backwards for refinable origins.
	 * Passes recursively through other connection chain elements.
	 * @return The origins or an empty list if the origins are not RefinableModels.
	 */
	public List<RefinableModel> listChainOrigins() {
		List<RefinableModel> origins = new ArrayList<RefinableModel>();
		List parents = listGrafParents();
		for (int i = 0; i < parents.size(); i++) {
			ModelWithConnections parent = (ModelWithConnections)parents.get(i);
			List<RefinableModel> parentOrigins = parent.listChainOrigins();
			// TODO: if (parentOrigins.size() == 0) { return new ArrayList; ? }
			origins.addAll(parentOrigins);
		}
		return origins;
	}

	public int getGrafDepth() {
		return findGrafDepth(new HashSet<ModelWithConnections>());
	}
	
	protected int findGrafDepth(Set<ModelWithConnections> visited) {
		// If a cycle is reached the depth is unknown.
		if (visited.contains(this)) {
			return UNKNOWN_DEPTH;
		}
		visited.add(this);
		
		int depth = UNKNOWN_DEPTH;
		List<ModelWithConnections> parents = listGrafParents();
		for (ModelWithConnections parent: parents) {
			int parentDepth = parent.findGrafDepth(visited);
			if (parentDepth != UNKNOWN_DEPTH) {
				depth = (depth != UNKNOWN_DEPTH) ? Math.min(depth, parentDepth + 1) : parentDepth + 1;
			}
		}
		return depth; 
	}

}
