package org.infrawebs.axiom.model;

import java.io.IOException;
import java.util.Set;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.PropertyDescriptor;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.ontology.RestorationException;
import org.infrawebs.ontology.shells.ParameterShell;
import org.omwg.ontology.Relation;
import org.wsmo.common.Identifier;

public class RelationModel extends ModelWithConnections
		implements Compactable, RefinableListHolder
{
	private static final long serialVersionUID = 200512242100L;

	public static final String ID_PROPERTY_RELATION_NAME = "ID_PROPERTY_RELATION_NAME";

	private transient Relation relation = null;
	private ParameterListModel parameterList = new ParameterListModel(this);
	private boolean isCompact = false;
	private Dimension expandedWidth = null;
	
	public RelationModel(AxiomContentsModel contents, Relation relation) {
		super(contents);
		this.relation = relation;

		// Add the relation's parameters.
		int count = relation.listParameters().size();
		for (byte i = 0; i < count; i++) {
			ParameterShell paramInfo = new ParameterShell(relation.getParameter(i));
			ParameterModel param = new ParameterModel(this.contents, paramInfo);
			this.parameterList.addChild(param);
		}

		// TODO: Is this necessary?
//		firePropertyChange(ID_PROPERTY_RELATION_NAME, null, relation);
	}

	public Relation getRelation() {
		return this.relation;
	}

	@Override
	public IPropertyDescriptor[] createPropertyDescriptors() {
		IPropertyDescriptor[] descriptors =	new IPropertyDescriptor[1];
		descriptors[0] = new PropertyDescriptor(ID_PROPERTY_RELATION_NAME, "Name");
		return descriptors;
	}

	@Override
	public Object getPropertyValue(Object id) {
		if (id.equals(ID_PROPERTY_RELATION_NAME)) {
			return AxiomEditorPlugin.ontoStore.getPrefixedName(this.relation);
		} 
		return null;
	}

	@Override
	public void setPropertyValue(Object id, Object value) {
		// No editable properties.
	}

	public ParameterListModel getRefinableList() {
		return this.parameterList;
	}

	@Override
	public void setConstraint(Object constraint) {
		Rectangle newRect = (Rectangle)constraint;
		if (!isCompact()) {
			// Keep models auto sized.
			newRect.height = -1;
		}
		else {
			// Keep original compact size.
			newRect.setSize(COMPACT_SIZE);			
		}
		super.setConstraint(newRect);
	}

	public boolean isCompact() {
		return this.isCompact;
	}

	public void setCompact(boolean isCompact) {
		boolean oldIsCompact = this.isCompact; 
		this.isCompact = isCompact;

		this.expandedWidth = MemberOfModel.recalculateConstraint(
				this, this.isCompact, this.expandedWidth);
		
		firePropertyChange(ID_PROPERTY_COMPACT,
				new Boolean(oldIsCompact), new Boolean(isCompact));
	}

	@Override
	protected int findGrafDepth(Set<ModelWithConnections> visited) {
		if (isFixedInExpression()) {
			return super.findGrafDepth(visited);
		}

		// This is a floating relation. Its depth is determined by its "deepest parameter".
		// This is the refining model after which the relation is included in the expression tree. 

		// If a cycle is reached the depth is unknown.
		if (visited.contains(this)) {
			return UNKNOWN_DEPTH;
		}
		visited.add(this);

		ParameterModel deepest = this.parameterList.getDeepestParameter();
		if (deepest != null) {
			ModelWithConnections parent = deepest.getRefiningModel(); 
			int parentDepth = parent.findGrafDepth(visited);
			if (parentDepth != UNKNOWN_DEPTH) {
				return parentDepth + 1;
			}
		}
		return UNKNOWN_DEPTH; 
	}

	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		// Write just the relation identifier, not the relation itself.
		out.writeObject(this.relation.getIdentifier());
	}

	private void readObject(java.io.ObjectInputStream in)
			throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		// Read the relation identifier.
		Identifier relationId = (Identifier)in.readObject();

		// Restore the definition of the relation.
		this.relation = (Relation)AxiomEditorPlugin.ontoStore.findDefinition(
				relationId, Relation.class);
		if (this.relation == null) {
			throw new RestorationException("Relation " + relationId + " not found.");
		}
	}

}
