package org.infrawebs.axiom.model;

import java.beans.PropertyChangeListener;

import org.eclipse.ui.views.properties.ComboBoxPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.TextPropertyDescriptor;
import org.infrawebs.common.Messages;

public class VarNameModel extends AbstractModel
{
	private static final long serialVersionUID = 200511101800L;

	public static class VarNameException extends Exception
	{
		private static final long serialVersionUID = 200509271500L;

		public VarNameException(String message) {
			super(message);
		}
	}
	
	public static final String ID_PROPERTY_VARNAME = "ID_PROPERTY_VARNAME";
	public static final String ID_PROPERTY_EXISTS_FLAG = "ID_PROPERTY_EXISTS_FLAG";
	private String name = null;
	private boolean existsFlag = false;
	private VarNameList varNameListModel = null;

	public VarNameModel(VarNameList varNameListModel, String prefix) {
		this.varNameListModel = varNameListModel;
		this.name = varNameListModel.generateUniqueVariableName(prefix);
	}

	public String getName() {
		return this.name;
	}

	// Changes the name preventing variable name duplication in the list.
	public void setName(String newName) throws VarNameException {
		if (newName.length() == 0) {
			throw new VarNameException("Cannot rename variable: the name is empty.");
		}
		String oldName = this.name;
		if (oldName.equals(newName)) {
			return;
		}

		// Prevent duplicate names.
		if (this.varNameListModel.get(newName) != null) {
			throw new VarNameException("Cannot rename variable: duplicate name \""
					+ newName + "\".");
		}
		this.varNameListModel.remove(oldName);
		this.name = newName;
		this.varNameListModel.add(this);
		firePropertyChange(ID_PROPERTY_VARNAME, oldName, newName);
	}

	public boolean getExistsFlag() {
		return this.existsFlag;
	}
	
	public void setExistsFlag(boolean existsFlag) {
		boolean oldExistsFlag = this.existsFlag;
		this.existsFlag = existsFlag;
		firePropertyChange(ID_PROPERTY_EXISTS_FLAG, new Boolean(oldExistsFlag), new Boolean(existsFlag));
	}

	@Override
	public IPropertyDescriptor[] createPropertyDescriptors() {
		IPropertyDescriptor[] descriptors = new IPropertyDescriptor[2];
		descriptors[0] = new TextPropertyDescriptor(ID_PROPERTY_VARNAME, "Name");
		descriptors[1] = new ComboBoxPropertyDescriptor(ID_PROPERTY_EXISTS_FLAG, "Exists", new String[] {"false", "true"});
		return descriptors;
	}

	@Override
	public Object getPropertyValue(Object id) {
		if (id.equals(ID_PROPERTY_VARNAME)) {
			return this.name;
		}
		else if (id.equals(ID_PROPERTY_EXISTS_FLAG)) {
			return new Integer(this.existsFlag ? 1 : 0);
		}
		return null;
	}

	@Override
	public void setPropertyValue(Object id, Object value) {
		if (id.equals(ID_PROPERTY_VARNAME)) {
			try {
				setName((String)value);
			}
			catch (VarNameException ex) {
				showWarning(ex);
			}
		}
		else if (id.equals(ID_PROPERTY_EXISTS_FLAG)) {
			int index = ((Integer)value).intValue();
			setExistsFlag(index == 1);
		}
	}
	
	// All models which contain this VarName are also its listeners.
	// When the first model interested in this object appears, the
	// VarName attaches itself to the list of variables.
	@Override
	public void addPropertyChangeListener(PropertyChangeListener listener) {
		if (!this.hasListeners()) {
			try {
				this.varNameListModel.add(this);
			}
			catch (VarNameException ex) {
				showWarning(ex);
			}
		}
		super.addPropertyChangeListener(listener);
	}

	// All models which contain this VarName are also its listeners.
	// When no more models are interested in this object, it detaches
	// itself from the list of variables.
	// Thus another VarName with the same name can be created later.
	@Override
	public void removePropertyChangeListener(PropertyChangeListener listener) {
		super.removePropertyChangeListener(listener);
		if (!this.hasListeners()) {
			this.varNameListModel.remove(this.name);
		}
	}

	public static void showWarning(VarNameException ex) {
		Messages.showWarning("Semantic inconsistency: " + ex.getMessage());
	}

}
