package org.infrawebs.axiom.operations;

import java.util.HashSet;
import java.util.List;

import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.commands.CommandStack;
import org.eclipse.gef.commands.CompoundCommand;
import org.infrawebs.axiom.editor.editparts.AxiomConnectionEditPart;
import org.infrawebs.axiom.model.AbstractModel;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.model.AxiomRootModel;
import org.infrawebs.axiom.model.ManhattanConnectionModel;
import org.infrawebs.axiom.model.MemberOfModel;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.OperatorModel;
import org.infrawebs.axiom.model.RefinableModel;
import org.infrawebs.axiom.model.RelationModel;
import org.infrawebs.axiom.model.VarNameHolderModel;
import org.infrawebs.axiom.model.VarNameModel;
import org.infrawebs.axiom.model.VarNameModel.VarNameException;
import org.infrawebs.axiom.model.commands.ConnectionCreateCommand;
import org.infrawebs.axiom.model.commands.ContainerAddChildCommand;
import org.infrawebs.axiom.model.commands.VarNameChangeCommand;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;

public class ModelOperations {

	public ModelOperations() {
		super();
	}
	
	public static void executeCommand(Command command) {
		CommandStack stack = (CommandStack)AxiomEditorPlugin.getAxiomEditor().getAdapter(CommandStack.class); 
		stack.execute(command);
	}
	
	public static Command getCreateCommand(ModelWithConnections model) {
		AxiomContentsModel contents = AxiomEditorPlugin.getAxiomContentsModel();
		if (contents != null && model != null) {
			return new ContainerAddChildCommand(contents, model);
		}
		return null;
	}
	
	public static Command getConnectionCommand(
			ModelWithConnections source, ModelWithConnections target) {
		ManhattanConnectionModel conn = new ManhattanConnectionModel();
		conn.setSource(source);
		conn.setTarget(target);
		Command command = new ConnectionCreateCommand(conn);
		return command;
	}

	/**
	 * Creates and completes a connection between two axiom elements.
	 */
	public static void connect(ModelWithConnections src, ModelWithConnections dst) {
		ManhattanConnectionModel conn = new ManhattanConnectionModel();
		conn.setSource(src);
		conn.setTarget(dst);
		conn.attachSource();
		conn.attachTarget();
	}

	/**
	 * Performs a DFS starting at the specified root and accumulates
	 * one VarNameChangeCommand for every element found which is
	 * associated with a variable different than varName.
	 */
	private static void getSpreadVarNameCommands(VarNameModel varName,
			ModelWithConnections root, CompoundCommand accumulator,
			HashSet<ModelWithConnections> visited)
			throws VarNameException {
		// End of recursion.
		if (visited.contains(root)) {
			return;
		}
		visited.add(root);

		// VarNameHolders change their own variable (if necessary).
		if (root instanceof VarNameHolderModel) {
			VarNameHolderModel holder = (VarNameHolderModel)root;
			if (holder.getVarNameModel() != varName) {
				Command changeCommand = new VarNameChangeCommand(holder, varName);
				accumulator.add(changeCommand);
			}
			// MemberOfModels get back to all their refinable origins
			// and start the change from there.
			if (root instanceof MemberOfModel) {
				for (RefinableModel origin: root.listChainOrigins()) {
					getSpreadVarNameCommands(varName, origin, accumulator, visited);
				}
			}
			// Refinables spread the change to their refining models.
			else if (root instanceof RefinableModel) {
				ModelWithConnections refining = ((RefinableModel)root).getRefiningModel();
				if (refining != null) {
					getSpreadVarNameCommands(varName, refining, accumulator, visited);
				}
			}
		}
		// Opereators spread the change to all their operands.
		else if (root instanceof OperatorModel) {
			for (ModelWithConnections operand: ((OperatorModel)root).listOperands()) {
				getSpreadVarNameCommands(varName, operand, accumulator, visited);
			}
		}
		// Relations spread nothing. They are just skipped.
		else if (root instanceof RelationModel) {
			//
		}
		// Any other elements abort the spreading.
		else {
			throw new VarNameException("Cannot spread variable ?" + varName.getName() +
					"\nto object of type " + root.getClass().getName() + ".");
		}
	}
	
	/**
	 * If a connection is created which is part of a connection chain
	 * originating from a refinable, then the VarNameModel of the chain's
	 * origin must be spread to all interested targets.
	 * 
	 * If the source is a refinable itself, it may not have a VarNameModel
	 * yet. In such a case a new VarNameModel is created which will be
	 * assigned to the refinable. 
	 */ 
	public static Command addSpreadVarNameCommands(Command connCommand,
			ModelWithConnections src, ModelWithConnections dst) {
		VarNameModel varNameModel = null;
		CompoundCommand accumulator = new CompoundCommand();
		if (src instanceof RefinableModel) {
			RefinableModel refinable = (RefinableModel)src;
			varNameModel = refinable.getVarNameModel();
			// If there is nothing to spread, create a new varNameModel.
			// Add a preliminary command to attach it to the refinable.
			if (varNameModel == null) {
				varNameModel = refinable.createVarNameModel();
				accumulator.add(new VarNameChangeCommand(refinable, varNameModel));
			}
		}
		else {
			List<RefinableModel> srcOrigins = src.listChainOrigins();
			// If the connection chain does not originate from refinables,
			// then no variables must be spread.
			if (srcOrigins.size() == 0) {
				return connCommand;
			}
			// All RefinableModels should hold the same VarNameModel,
			// so we take the VarNameModel from the first RefinableModel. 
			varNameModel = srcOrigins.get(0).getVarNameModel();
		}

		// Start the list of commands with the connection command itself.
		accumulator.add(connCommand);

		try {
			getSpreadVarNameCommands(varNameModel, dst, accumulator, new HashSet<ModelWithConnections>());
			return accumulator;
		}
		catch (VarNameException ex) {
			// Cancel the connection command because the variable cannot be spread.
			System.out.println(ex.getMessage());
			return null;
		}
	}

	public static void createAndConnect(
			ModelWithConnections newModel,
			ModelWithConnections existingSourceModel,
			ModelWithConnections targetModel // Usually same as the newModel.
	) {
		CompoundCommand compoundCommand = new CompoundCommand();		
		compoundCommand.add(getCreateCommand(newModel));
		Command connCommand = getConnectionCommand(existingSourceModel, targetModel);
		connCommand = addSpreadVarNameCommands(
				connCommand, existingSourceModel, targetModel);
		compoundCommand.add(connCommand);
		executeCommand(compoundCommand);
	}

	public static void createAndConnectToAxiomRoot(ModelWithConnections model) {
		// Find the axiom root element.
		AxiomContentsModel contents = AxiomEditorPlugin.getAxiomContentsModel();
		AxiomRootModel axiomRoot = contents.getAxiomRoot();

		// If the axiom root is not connected, connect it to the new element.
		if (axiomRoot.getTopModel() == null) {
			createAndConnect(model, axiomRoot, model);
		}
		else {
			executeCommand(getCreateCommand(model));
		}
	}

	public static Rectangle getMiddleOf(AbstractModel m1, AbstractModel m2) {
		Rectangle r1 = (Rectangle)m1.getConstraint();
		Rectangle r2 = (Rectangle)m2.getConstraint();
		
		int mid_x = ( r1.x + r2.x ) / 2;
		int mid_y = ( r1.y + r2.y ) / 2;
		return new Rectangle(mid_x, mid_y, -1, -1);		
	}

	public static Rectangle getMiddleOf(AxiomConnectionEditPart connectionEditPart) {
		Point p = connectionEditPart.getConnectionFigure().getBounds().getCenter(); 
		return new Rectangle(p.x - 50, p.y - 15, -1, -1);		
	}

}
