package org.infrawebs.axiom.operations;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IPath;
import org.infrawebs.axiom.logexpr.LEShell;
import org.infrawebs.axiom.logexpr.WsmoLogExprImport;
import org.infrawebs.axiom.logexpr.WsmoTopEntityExport;
import org.infrawebs.axiom.logexpr.WsmoTopEntityImport;
import org.infrawebs.axiom.model.AxiomContentsModel;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Messages;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.RestorationException;
import org.infrawebs.ontology.shells.OntologyShell;
import org.omwg.ontology.Ontology;
import org.wsmo.common.IRI;
import org.wsmo.common.TopEntity;
import org.wsmostudio.runtime.WSMORuntime;

public class Persistence
{
	private static WsmoTopEntityExport entityExport = new WsmoTopEntityExport();
	private static WsmoTopEntityImport wsmoTopEntityImport = new WsmoTopEntityImport();
	
	/**
	 * Serializes an axiom model to an *.axiom file.
	 */
	public boolean save(Set<OntologyShell> usedOntologies,
			AxiomContentsModel contents, String fileName) {
		try {
			FileOutputStream fs = new FileOutputStream(fileName);
			ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(fs));
			saveUsedOntologies(out, usedOntologies);
			out.writeObject(contents);
			out.close();
			return true;
		}
		catch (IOException ex) {
			Messages.showError("Error saving file \"" + fileName + "\".\n" +
					"\nMessage: " + ex.getMessage() + "\n" +
					"Exception: " + ex.getClass().getName());
			return false;
		}
	}

	/**
	 * Serializes the identifiers of all ontologies used by the axiom model. 
	 */
	private void saveUsedOntologies(ObjectOutputStream out,
			Set<OntologyShell> usedOntologies) throws IOException {
		List<IRI> ontologyIds = new ArrayList<IRI>();
		for (OntologyShell ontology: usedOntologies) {
			ontologyIds.add((IRI)ontology.getOntology().getIdentifier());
		}
		out.writeObject(ontologyIds);
	}

	/**
	 * Loads an axiom model from a file. The process could be interactive.
	 */
	public AxiomContentsModel load(IPath path) {
		if (path == null) {
			return null;
		}

		if (path.getFileExtension().equals("wsml")) {
			List selected = selectLogicalExpressions(path.toString());
			if (selected == null) {
				return null;  // User abort.
			}
			LEShell leShell = (LEShell)selected.get(0);
			return WsmoLogExprImport.createAxiomModel(
					leShell.getLogicalExpression(),
					leShell.getAxiomName(),
					leShell.listImportedOntos());
		}
		// The extension of the file is not "wsml". Try to deserialize a binary file.
		return deserializeAxiomModel(path.toString());
	}

	/**
	 * Parses a WSML file, extracts all logical expressions from it and asks the user to select one.
	 * @return The first logical expression selected by the user or null on error.
	 */
	private List selectLogicalExpressions(String fileName) {
		// PATCH: This forces the reset of the dummy web service and the dummy capability
		// which are used when exporting axioms. They are cached by the WSMO factory and
		// their contents needs to be cleared before loading new contents into them.
		entityExport.createWebService(null, null);
				
		TopEntity[] tops = null;
		try {
			tops = Util.parseFile(fileName);
		}
		catch (Exception ex) {
			Messages.showError("Error loading file " + fileName + "\n" + ex.getMessage());
			return null;
		}
		
		List<LEShell> leShells = wsmoTopEntityImport.listLEShells(tops);
		if (leShells.isEmpty()) {
			Messages.showWarning("No logical expressions found in file " + fileName);
			return null;
		}
		// If there is a single logical expression in the file, select it.
		if (leShells.size() == 1) {
			return leShells;
		}
		// If there are multiple logical expressions in the file, let the user select some.
		return wsmoTopEntityImport.selectLogicalExpressions(leShells);
	}

	/**
	 * Deserializes an axiom model from an *.axiom file.
	 */
	@SuppressWarnings("unchecked")
	private AxiomContentsModel deserializeAxiomModel(String fileName) {
		try {
			FileInputStream fs = new FileInputStream(fileName);
			ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(fs));
			List<IRI> ontologyIds = (List<IRI>)in.readObject();
			loadRequiredOntologies(ontologyIds);
			AxiomContentsModel contents = (AxiomContentsModel)in.readObject();
			in.close();
			return contents;
		}
		catch (ClassNotFoundException ex) {
			Messages.showWarning("Error loading file \"" + fileName + "\".\n" +
					"A required class is missing.\n" +
					"\nDetails: " + ex.getMessage());
		}
		catch (InvalidClassException ex) {
			Messages.showWarning("File format incompatible.\n" +
					"The file \"" + fileName + "\" was probably created with a different version of AxiomEditor.\n" +
					"\nDetails: " + ex.getMessage());
		}
		catch (IOException ex) {
			Messages.showWarning("Error loading file \"" + fileName + "\".\n" +
					"\nDetails: " + ex.getMessage() + "\n" +
					"Exception: " + ex.getClass().getName());
		}
		return null;
	}

	/**
	 * Ensures that all required ontologies are loaded before the axiom model is deserialized. 
	 */
	private void loadRequiredOntologies(List<IRI> ontologyIds) throws IOException {
		if (ontologyIds == null) {
			return;
		}
		for (IRI iri: ontologyIds) {
			if (AxiomEditorPlugin.ontoStore.getOntology(iri) != null) {
				continue;  // Ontology already present in OntoStore.
			}
			Ontology ontology = WSMORuntime.getIOManager().doLoadOntology(iri);
			if (ontology != null) {
				AxiomEditorPlugin.ontoStore.addOntology(new OntologyShell(ontology));
			}
			else {
				throw new RestorationException("Required ontology missing: " + iri);
			}
		}
	}

}
