package org.infrawebs.axiom.operations;

import java.util.List;

import org.infrawebs.axiom.model.AxiomRootModel;
import org.infrawebs.axiom.model.MemberOfModel;
import org.infrawebs.axiom.model.ModelWithConnections;
import org.infrawebs.axiom.model.OperatorModel;
import org.infrawebs.axiom.model.RefinableModel;
import org.infrawebs.axiom.model.RelationModel;
import org.infrawebs.axiom.plugin.AxiomEditorPlugin;
import org.infrawebs.common.Util;
import org.infrawebs.ontology.OntoFilter;
import org.omwg.ontology.Concept;
import org.omwg.ontology.Type;

public class SemanticConsistence {
	private static OntoFilter ontoFilter = new OntoFilter(false);

	public static boolean allowConnection(ModelWithConnections src, ModelWithConnections dst) {
		// Do not allow duplicate or bidirectional connections.
		if (src.listGrafChildren().contains(dst) || dst.listGrafChildren().contains(src)) {
			return false;
		}

		List<RefinableModel> chainOrigins = src.listChainOrigins();
		if (chainOrigins.size() > 0) {
			return	dst instanceof MemberOfModel &&
					areCompatibleRefinablesWithMemberOf(chainOrigins, (MemberOfModel)dst) ||
					dst instanceof OperatorModel &&
					areCompatibleRefinablesWithSubExpression(chainOrigins, (OperatorModel)dst);				
		}
		return	dst instanceof MemberOfModel ||
				dst instanceof RelationModel ||				
				dst instanceof OperatorModel;
	}

	/**
	 * @return True if ALL refinables (from the chainOrigins list) are
	 * compatible with the memberOfModel.
	 */
	private static boolean areCompatibleRefinablesWithMemberOf(
			List<RefinableModel> chainOrigins, MemberOfModel memberOfModel) {
		for (RefinableModel refinable: chainOrigins) {
			ontoFilter.baseTypes = refinable.getRefinable().listTypes();
			if (!ontoFilter.isCompatible(memberOfModel.getConcept())) {
				return false;
			}
		}
		return true; 
	}

	/**
	 * Performs the areCompatibleRefinablesWithMemberOf check recursively.
	 */
	private static boolean areCompatibleRefinablesWithSubExpression(
			List<RefinableModel> chainOrigins, OperatorModel operatorModel) {
		for (Object operand: operatorModel.listOperands()) {
			// Check MemberOf for compatibility.
			if (operand instanceof MemberOfModel &&
					areCompatibleRefinablesWithMemberOf(
					chainOrigins, (MemberOfModel)operand)) {
				continue;
			}
			// Check all operands for compatibility.
			if (operand instanceof OperatorModel &&
					areCompatibleRefinablesWithSubExpression(
					chainOrigins, (OperatorModel)operand)) {
				continue;
			}
			return false;
		}
		return true;
	}

	public static boolean allowMoreChildren(ModelWithConnections model) {
		int count = model.listSourceConnections().size();
		return (model instanceof AxiomRootModel && (count == 0))
				|| (model instanceof MemberOfModel && false)				
				|| (model instanceof RelationModel && false)				
				|| (model instanceof RefinableModel && (count == 0))
				|| (model instanceof OperatorModel && allowMoreOperands((OperatorModel)model));
	}

	public static boolean allowMoreParents(ModelWithConnections model) {
		int count = model.listTargetConnections().size();
		return (model instanceof AxiomRootModel && false)
				|| (model instanceof MemberOfModel && true/*(count == 0)*/)				
				|| (model instanceof RelationModel && true/*(count == 0)*/)				
				|| (model instanceof RefinableModel && false)
				|| (model instanceof OperatorModel && (count == 0));
	}

	public static boolean allowMoreOperands(OperatorModel operator) {
		return ((operator.getOperatorType() == OperatorModel.OPERATOR_NOT) && (operator.listOperands().size() < 1))
				|| ((operator.getOperatorType() == OperatorModel.OPERATOR_AND) && true)
				|| ((operator.getOperatorType() == OperatorModel.OPERATOR_OR) && true);				
	}

	public static boolean isValueType(RefinableModel refinableModel) {
		if (refinableModel == null) {
			return false;
		}
		Type firstType = refinableModel.getRefinable().getFirstType();
		Concept concept = AxiomEditorPlugin.ontoStore.liftTypeToConcept(firstType);
		return Util.isValueType(concept);
	}
}
