package org.infrawebs.common;

import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import org.eclipse.draw2d.geometry.Point;
import org.eclipse.swt.widgets.TreeItem;
import org.omwg.logicalexpression.Constants;
import org.omwg.ontology.Concept;
import org.omwg.ontology.Ontology;
import org.omwg.ontology.OntologyElement;
import org.omwg.ontology.Type;
import org.omwg.ontology.WsmlDataType;
import org.wsmo.common.Entity;
import org.wsmo.common.IRI;
import org.wsmo.common.TopEntity;
import org.wsmo.common.exception.InvalidModelException;
import org.wsmo.common.exception.SynchronisationException;
import org.wsmo.wsml.ParserException;
import org.wsmostudio.runtime.WSMORuntime;

public final class Util
{
	public static final String LOCAL_NAME_DELIMITER = Constants.SQNAME_DEL;

	public static String getLocalName(Object o) {
		if (o == null) {
			throw new IllegalArgumentException();
		}
		if (o instanceof Entity) {
			Entity entity = (Entity)o;
			return getLocalName(entity.getIdentifier());
		}
		else if (o instanceof WsmlDataType) {
			WsmlDataType type = (WsmlDataType)o;
			return getLocalName(type.getIRI());
		}
		else if (o instanceof IRI) {
			IRI iri = (IRI)o;
			return iri.getLocalName();
		}
		else {
			IRI iri = WSMORuntime.getRuntime().getWsmoFactory().createIRI(o.toString());
			return getLocalName(iri);
		}
	}

	public static String getNamespace(String id) {
		IRI iri = WSMORuntime.getRuntime().getWsmoFactory().createIRI(id);
		String ns = iri.getNamespace();
		// Strip trailing delimiter.
		if (ns != null && ns.endsWith(LOCAL_NAME_DELIMITER)) {
			ns = ns.substring(0, ns.length() - LOCAL_NAME_DELIMITER.length());
		}
		return ns;
	}
	
	public static String getNamespace(OntologyElement element) {
		if (element == null) {
			throw new IllegalArgumentException();
		}
		return getNamespace(element.getIdentifier().toString());
	}

	// Compares types by their identifiers. Thus a built-in type and a concept might prove equal.
	public static boolean equalTypes(Type type1, Type type2) {
		Object id1 = (type1 instanceof WsmlDataType)
				? ((WsmlDataType)type1).getIRI()
				: ((Concept)type1).getIdentifier();
		Object id2 = (type2 instanceof WsmlDataType)
				? ((WsmlDataType)type2).getIRI()
				: ((Concept)type2).getIdentifier();
		return id1.equals(id2);
	}
	
	public static boolean isSubConcept(Concept subConcept, Type type) {
		Iterator i = subConcept.listSuperConcepts().iterator();
		while (i.hasNext()) {
			Concept superConcept = (Concept)i.next();
			if (equalTypes(superConcept, type)) {
				return true;
			}
			if (isSubConcept(superConcept, type)) {
				return true;
			}
		}
		return false;
	}

	public static boolean isWsmlDataType(Concept concept, String dataTypeName) {
		return concept.getIdentifier().toString().equals(Constants.WSML_NAMESPACE + dataTypeName);
	}

	public static boolean isWsmlSyntaxOntology(Ontology ontology) {
		String wsml = Constants.WSML_NAMESPACE.substring(0, Constants.WSML_NAMESPACE.length() - 1);
		return ontology.getIdentifier().toString().equals(wsml);
	}
	
	/**
	 * Checks if it is equal to or a subConcept of a Built-in WSML type.
	 */
	public static boolean isValueType(Concept subConcept) {
		// subConcept might be a Proxy class which connot call getOntology().
		Ontology o = null;
		try {
			o = subConcept.getOntology();
		}
		catch (SynchronisationException ex) {
			// The subConcept is a proxy and o remains null.
		}

		if (o != null && isWsmlSyntaxOntology(o)) {
			return true;
		}
		// probably just the first superConcept is enough, no need to iterate...
		Iterator i = subConcept.listSuperConcepts().iterator();
		while (i.hasNext()) {
			Concept superConcept = (Concept)i.next();
			if (isValueType(superConcept))
				return true;
		}
		return false;
	}	

	// TODO: Move this utility method somewhere else.
	public static void ExpandFullyTreeItem(TreeItem ti, boolean expanded) {
		ti.setExpanded(expanded);
		TreeItem[] items = ti.getItems();
		for (int i = 0; i < items.length; i++) {
			ExpandFullyTreeItem(items[i], expanded);
		}
	}

	public static Point getRandomLocation() {
		Random rnd = new Random();
		return new Point(rnd.nextInt(400), rnd.nextInt(300));
	}
	
	public static TopEntity[] parseFile(String fileName) throws IOException, ParserException, InvalidModelException {
		FileReader reader = new FileReader(fileName);
		try {
	    	return WSMORuntime.getRuntime().getWsmlParser().parse(reader);
		}
		finally {
			reader.close();
		}
	}

	public static String serialize(TopEntity topEntity) {
		if (topEntity == null) {
			return "";
		}
		TopEntity[] topEntities = new TopEntity[] { topEntity };
		StringBuffer buffer = new StringBuffer();
		WSMORuntime.getRuntime().getWsmlSerializer().serialize(topEntities, buffer);
		return buffer.toString();
	}

	// Converts a Set of Ontologies to a List<Ontology>.
	public static List<Ontology> listImportedOntos(TopEntity top) {
		List<Ontology> importedOntos = new ArrayList<Ontology>();
		Iterator i = top.listOntologies().iterator();
		while (i.hasNext()) {
			importedOntos.add((Ontology)i.next());
		}
		return importedOntos;
	}
}
