package org.infrawebs.ontology;

import java.util.Iterator;
import java.util.Set;

import org.infrawebs.common.Util;
import org.infrawebs.ontology.shells.ParameterShell;
import org.omwg.ontology.Concept;
import org.omwg.ontology.Instance;
import org.omwg.ontology.Parameter;
import org.omwg.ontology.Relation;
import org.omwg.ontology.Type;

public class OntoFilter {
	public Set<Type> baseTypes = null;

	public boolean showMetadata = false;
	public boolean showConcepts = false;
	public boolean showAttributes = false;
	public boolean showInstances = false;
	public boolean showRelations = false;

	public boolean selectConcepts = false;
	public boolean selectInstances = false;
	public boolean selectRelations = false;
	public boolean selectParameters = false;

	public OntoFilter(boolean acceptAll) {
		this.showMetadata = acceptAll;
		this.showConcepts = acceptAll;
		this.showAttributes = acceptAll;
		this.showInstances = acceptAll;
		this.showRelations = acceptAll;
		this.selectConcepts = acceptAll;
		this.selectInstances = acceptAll;
		this.selectRelations = acceptAll;
		this.selectParameters = acceptAll;
	}
	
	// Returns true if the concept is equal to or is a subconcept
	// of any of the base types.
	public boolean isCompatible(Type type) {
		if (this.baseTypes == null) {
			return true;
		}
		Iterator basei = this.baseTypes.iterator();
		while (basei.hasNext()) {
			Type baseType = (Type)basei.next();
			if (Util.equalTypes(type, baseType) ||
					(type instanceof Concept &&
					 Util.isSubConcept((Concept)type, baseType))) {
				return true;
			}
		}
		return false;
	}

	// Returns true if the concept is equal to or is a super concept
	// of any of the base types.
	public boolean isReverseCompatible(Type type) {
		if (this.baseTypes == null) {
			return true;
		}
		for (Type baseType: this.baseTypes) {
			if (Util.equalTypes(type, baseType) ||
					(baseType instanceof Concept &&
					 Util.isSubConcept((Concept)baseType, type))) {
				return true;
			}
		}
		return false;
	}

	// An instance is compatible if any of the concepts it belongs to is compatible.
	public boolean isCompatible(Instance instance) {
		Iterator i = instance.listConcepts().iterator();
		while (i.hasNext()) {
			Concept concept = (Concept)i.next();
			if (isCompatible(concept)) {
				return true;
			}
		}
		return false;
	}
	
	// A relation is compatible if any of its parameters is compatible.
	public boolean isCompatible(Relation relation) {
		Iterator i = relation.listParameters().iterator();
		while (i.hasNext()) {
			Parameter parameter = (Parameter)i.next();
			if (isCompatible(parameter)) {
				return true;
			}
		}
		return false;
	}
	
	// A parameter is compatible if any of the types it belongs to
	// is a super concept of at least one base type.
	public boolean isCompatible(Parameter parameter) {
		Iterator i = parameter.listTypes().iterator();
		while (i.hasNext()) {
			Type type = (Type)i.next();
			if (isReverseCompatible(type)) {
				return true;
			}
		}
		return false;
	}
	
	public boolean allowSelect(Object object) {
		return (object instanceof Concept && this.selectConcepts && isCompatible((Concept)object)) ||
				(object instanceof Instance && this.selectInstances && isCompatible((Instance)object)) ||
				(object instanceof Relation && this.selectRelations && isCompatible((Relation)object)) ||
				(object instanceof ParameterShell && this.selectParameters && isCompatible(((ParameterShell)object).getParameter()));
	}
	
}
