package org.infrawebs.ontology;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.infrawebs.common.Util;
import org.infrawebs.ontology.shells.OntologyShell;
import org.omwg.logicalexpression.Constants;
import org.omwg.ontology.Concept;
import org.omwg.ontology.Ontology;
import org.omwg.ontology.OntologyElement;
import org.omwg.ontology.Type;
import org.omwg.ontology.WsmlDataType;
import org.wsmo.common.IRI;
import org.wsmo.common.Identifier;
import org.wsmo.common.exception.SynchronisationException;

public class OntoStore {
	public static String ADD_ONTOLOGY = "ADD_ONTOLOGY";

	public static String JUMP_TO_ELEMENT = "LOCATE_ONTOLOGY_ELEMENT";

	private HashMap<Identifier, OntologyShell> ontologiesById = new HashMap<Identifier, OntologyShell>();
	private HashSet<String> ontologyPrefixes = new HashSet<String>();
	private PropertyChangeSupport listeners = new PropertyChangeSupport(this);

	public OntologyShell[] listOntologies() {
		OntologyShell[] ontoArray = new OntologyShell[this.ontologiesById.size()];
		Iterator iter = this.ontologiesById.values().iterator();
		int i = 0;
		while (iter.hasNext()) {
			ontoArray[i++] = (OntologyShell) iter.next();
		}
		return ontoArray;
	}

	public void addOntology(OntologyShell ontologyShell) {
		if (ontologyShell == null) {
			throw new IllegalArgumentException(
					"Trying to add a null ontology to the OntoStore.");
		}
		String prefix = generateUniquePrefix(ontologyShell);
		ontologyShell.setPrefix(prefix);

		this.ontologiesById.put(ontologyShell.getOntology().getIdentifier(),
				ontologyShell);
		this.ontologyPrefixes.add(ontologyShell.getPrefix());

		this.listeners.firePropertyChange(ADD_ONTOLOGY, null, ontologyShell);
	}

	// Returns the ontology with the identifier specified.
	public OntologyShell getOntology(Identifier ontoId) {
		return this.ontologiesById.get(ontoId);
	}

	// Returns the ontology which is the owner of the object.
	// NOTE: Many ontologies may own elements with the same identifier
	// but only one ontology owns this particular object.
	public OntologyShell getOntologyContaining(OntologyElement element) {
		if (element == null) {
			throw new IllegalArgumentException(
					"Trying to find the ontology of a null element.");
		}
		try {
			// There might be no owner ontology (e.g. the type of an
			// attribute may not be owned by the owner of the attribute).
			Ontology o = element.getOntology();
			return (o != null) ? getOntology(o.getIdentifier()) : null;
		} catch (SynchronisationException ex) {
			return null;
		}
	}

	// Returns the object containing the full definition of the element.
	// NOTE: Many ontologies may reference the same identifier but at
	// most one ontology defines the structure behind that identifier.
	public OntologyElement findDefinition(OntologyElement element) {
		if (element == null) {
			throw new IllegalArgumentException(
					"Trying to find the definition of a null element.");
		}
		for (OntologyShell ao: this.ontologiesById.values()) {
			OntologyElement definition = ao.findDefinition(element);
			if (definition != null) {
				return definition;
			}
		}
		return null;
	}

	// Works for entities only. Doesn't work for attributes/parameters.
	public OntologyElement findDefinition(Identifier id, Class elementClass) {
		if (id == null) {
			throw new IllegalArgumentException(
					"Trying to find the definition of a null identifier.");
		}
		for (OntologyShell ao: this.ontologiesById.values()) {
			OntologyElement definition = ao.findDefinition(id, elementClass);
			if (definition != null) {
				return definition;
			}
		}
		return null;
	}

	// Returns the local name of the element, always prefixed:
	// - either by the local prefix of the ontology that defines it
	// - or by the namespace prefix of its imported ontology
	public String getPrefixedName(Object o) {
		if (o == null) {
			throw new IllegalArgumentException(
					"Trying to get the prefix of a null object.");
		}

		// Try to find the ontology defining the object and use its prefix.
		OntologyElement definition = null;
		if (o instanceof OntologyElement) {
			OntologyElement element = (OntologyElement) o;
			definition = findDefinition(element);
		}
		if (o instanceof WsmlDataType) {
			WsmlDataType wsmlDataType = (WsmlDataType) o;
			definition = findDefinition(wsmlDataType.getIRI(), Concept.class);
		}
		if (definition != null) {
			OntologyShell ao = getOntologyContaining(definition);
			if (ao != null) {
				if (Util.isWsmlSyntaxOntology(ao.getOntology())) {
					return "_" + Util.getLocalName(o);
				}
				return ao.getPrefix() + Constants.SQNAME_DEL + Util.getLocalName(o);
			}
		}

		// If the definition was not found try to apply the local namespace
		// prefix.
		if (o instanceof OntologyElement) {
			OntologyElement element = (OntologyElement) o;
			OntologyShell ao = getOntologyContaining(element);
			if (ao != null) {
				return ao.getQualifiedName(o);
			}
		}

		// If the method did not return until now, as a last resort use the
		// local name.
		return Util.getLocalName(o);
	}

	// Returns a coma separated list of prefixed names.
	public String getPrefixedNames(Set elements) {
		StringBuffer result = new StringBuffer();
		Iterator i = elements.iterator();
		while (i.hasNext()) {
			if (result.length() > 0) {
				result.append(", ");
			}
			result.append(getPrefixedName(i.next()));
		}
		return result.toString();
	}

	/**
	 * Tries to find the element in the ontology that defines an identifier.
	 * Makes the listeners select that element in their views.
	 * @return false if the defining ontology was not found in the store.
	 */
	OntologyElement jumpToDefinition(OntologyElement element) {
		OntologyElement definition = findDefinition(element);
		if (definition != null) {
			this.listeners.firePropertyChange(JUMP_TO_ELEMENT, null, definition);
			return definition;
		}
		this.listeners.firePropertyChange(JUMP_TO_ELEMENT, null, element);
		return null;
	}

	public void addListener(PropertyChangeListener listener) {
		this.listeners.addPropertyChangeListener(listener);
	}

	public void removeListener(PropertyChangeListener listener) {
		this.listeners.removePropertyChangeListener(listener);
	}

	// Appends the smallest possible number to the last part of the
	// ontology identifier so that name duplication is avoided.
	private String generateUniquePrefix(OntologyShell ontologyShell) {
		// Usually the local name can be extracted directly.
		String localName = Util.getLocalName(ontologyShell.getOntology());

		// Sometimes identifiers have symbols at the end which obscure the
		// local name. In such cases try to extract the local name manually.
		// Get the alpha-numeric charactars after the last slash.
		if (localName.length() == 0) {
			String id = ontologyShell.getOntology().getIdentifier().toString();
			int lastSlash = id.lastIndexOf('/');
			if (lastSlash >= 0) {
				int lastChar = lastSlash + 1;
				while (lastChar < id.length()
						&& Character.isLetterOrDigit(id.charAt(lastChar))) {
					lastChar++;
				}
				localName = id.substring(lastSlash + 1, lastChar);
			}
		}

		// As a last resort, use a constant prefix.
		if (localName.length() == 0) {
			localName = "ontology";
		}

		if (!this.ontologyPrefixes.contains(localName))
			return localName;

		int num = 1;
		while (this.ontologyPrefixes.contains(localName + num)) {
			num++;
		}
		return localName + num;
	}

	/**
	 * If the type is a built-in WSML data type, lifts it to a concept from the
	 * system ontology "wsml-syntax" which should always be loaded. If the type
	 * is a concept, returns it unchanged. If the system ontology is missing
	 * returns null.
	 */
	public Concept liftTypeToConcept(Type type) {
		if (type == null) {
			throw new IllegalArgumentException("Trying to lift a null type.");
		}
		if (type instanceof WsmlDataType) {
			IRI iri = ((WsmlDataType) type).getIRI();
			OntologyElement element = findDefinition(iri, Concept.class);
			if (element instanceof Concept) {
				return (Concept)element;
			}
		}
		else if (type instanceof Concept) {
			return (Concept)type;
		}
		return null;
	}

}
